'''
NOTES:
Includes:
Takeoff, Climb, Cruise, Descent, Landing
Computed Aero
N3CC data
'''
import unittest

import dymos as dm
import numpy as np
import openmdao.api as om
import scipy.constants as _units
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs
from openmdao.utils.testing_utils import require_pyoptsparse

from aviary.mission.flops_based.phases.build_landing import Landing
from aviary.mission.flops_based.phases.build_takeoff import Takeoff
from aviary.utils.functions import set_aviary_initial_values
from aviary.utils.test_utils.assert_utils import warn_timeseries_near_equal
from aviary.validation_cases.validation_tests import get_flops_inputs
from aviary.variable_info.functions import setup_trajectory_params
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables_in import VariablesIn
from aviary.mission.flops_based.phases.energy_phase import EnergyPhase

from aviary.variable_info.variables import Aircraft, Dynamic, Mission
from aviary.subsystems.premission import CorePreMission
from aviary.interface.default_phase_info.height_energy import default_premission_subsystems, default_mission_subsystems
from aviary.utils.preprocessors import preprocess_crewpayload

try:
    import pyoptsparse
except ImportError:
    pyoptsparse = None


# benchmark for simple sizing problem on the N3CC
def run_trajectory(sim=True):
    prob = om.Problem(model=om.Group())
    if pyoptsparse:
        driver = prob.driver = om.pyOptSparseDriver()
        driver.options["optimizer"] = "SNOPT"
        # driver.declare_coloring()  # currently disabled pending resolve of issue 2507
        if driver.options["optimizer"] == "SNOPT":
            driver.opt_settings["Major iterations limit"] = 45
            driver.opt_settings["Major optimality tolerance"] = 1e-4
            driver.opt_settings["Major feasibility tolerance"] = 1e-6
            driver.opt_settings["iSumm"] = 6
        elif driver.options["optimizer"] == "IPOPT":
            driver.opt_settings["max_iter"] = 100
            driver.opt_settings["tol"] = 1e-3
            driver.opt_settings['print_level'] = 4

    else:
        driver = prob.driver = om.ScipyOptimizeDriver()
        opt_settings = prob.driver.opt_settings

        driver.options['optimizer'] = 'SLSQP'
        opt_settings['maxiter'] = 100
        opt_settings['ftol'] = 5.0e-3
        opt_settings['eps'] = 1e-2

    ##########################################
    # Aircraft Input Variables and Options   #
    ##########################################

    aviary_inputs = get_flops_inputs('N3CC')

    aviary_inputs.set_val(Mission.Landing.LIFT_COEFFICIENT_MAX,
                          2.4, units="unitless")
    aviary_inputs.set_val(Mission.Takeoff.LIFT_COEFFICIENT_MAX,
                          2.0, units="unitless")
    aviary_inputs.set_val(
        Mission.Takeoff.ROLLING_FRICTION_COEFFICIENT,
        val=.0175, units="unitless")

    takeoff_fuel_burned = 577  # lbm TODO: where should this get connected from?
    takeoff_thrust_per_eng = 24555.5  # lbf TODO: where should this get connected from?
    takeoff_L_over_D = 17.35  # TODO: should this come from aero?

    aviary_inputs.set_val(Mission.Takeoff.FUEL_SIMPLE,
                          takeoff_fuel_burned, units='lbm')
    aviary_inputs.set_val(Mission.Takeoff.LIFT_OVER_DRAG,
                          takeoff_L_over_D, units="unitless")
    aviary_inputs.set_val(Mission.Design.THRUST_TAKEOFF_PER_ENG,
                          takeoff_thrust_per_eng, units='lbf')

    alt_airport = 0  # ft

    alt_i_climb = 0*_units.foot  # m
    alt_f_climb = 35000.0*_units.foot  # m
    mass_i_climb = 131000*_units.lb  # kg
    mass_f_climb = 126000*_units.lb  # kg
    # initial mach set to lower value so it can intersect with takeoff end mach
    # mach_i_climb = 0.3
    mach_i_climb = 0.2
    mach_f_climb = 0.79
    distance_i_climb = 0*_units.nautical_mile  # m
    distance_f_climb = 160.3*_units.nautical_mile  # m
    t_i_climb = 2 * _units.minute  # sec
    t_f_climb = 26.20*_units.minute  # sec
    t_duration_climb = t_f_climb - t_i_climb

    alt_i_cruise = 35000*_units.foot  # m
    alt_f_cruise = 35000*_units.foot  # m
    alt_min_cruise = 35000*_units.foot  # m
    alt_max_cruise = 35000*_units.foot  # m
    mass_i_cruise = 126000*_units.lb  # kg
    mass_f_cruise = 102000*_units.lb  # kg
    cruise_mach = 0.79
    distance_i_cruise = 160.3*_units.nautical_mile  # m
    distance_f_cruise = 3243.9*_units.nautical_mile  # m
    t_i_cruise = 26.20*_units.minute  # sec
    t_f_cruise = 432.38*_units.minute  # sec
    t_duration_cruise = t_f_cruise - t_i_cruise

    alt_i_descent = 35000*_units.foot
    # final altitude set to 35 to ensure landing is feasible point
    # alt_f_descent = 0*_units.foot
    alt_f_descent = 35*_units.foot
    mach_i_descent = 0.79
    mach_f_descent = 0.3
    mass_i_descent = 102000*_units.pound
    mass_f_descent = 101000*_units.pound
    distance_i_descent = 3243.9*_units.nautical_mile
    distance_f_descent = 3378.7*_units.nautical_mile
    t_i_descent = 432.38*_units.minute
    t_f_descent = 461.62*_units.minute
    t_duration_descent = t_f_descent - t_i_descent

    ##########################
    # Design Variables       #
    ##########################

    # Nudge it a bit off the correct answer to verify that the optimize takes us there.
    aviary_inputs.set_val(Mission.Design.GROSS_MASS, 135000.0, units='lbm')

    prob.model.add_design_var(Mission.Design.GROSS_MASS, units='lbm',
                              lower=100000.0, upper=200000.0, ref=135000)

    takeoff_options = Takeoff(
        airport_altitude=alt_airport,  # ft
        # no units
        num_engines=aviary_inputs.get_val(Aircraft.Engine.NUM_ENGINES)
    )

    ##################
    # Define Phases  #
    ##################
    num_segments_climb = 6
    num_segments_cruise = 1
    num_segments_descent = 5

    climb_seg_ends, _ = dm.utils.lgl.lgl(num_segments_climb + 1)
    descent_seg_ends, _ = dm.utils.lgl.lgl(num_segments_descent + 1)

    transcription_climb = dm.Radau(
        num_segments=num_segments_climb, order=3, compressed=True,
        segment_ends=climb_seg_ends)
    transcription_cruise = dm.Radau(
        num_segments=num_segments_cruise, order=3, compressed=True)
    transcription_descent = dm.Radau(
        num_segments=num_segments_descent, order=3, compressed=True,
        segment_ends=descent_seg_ends)

    climb_options = EnergyPhase(
        'test_climb',
        user_options=AviaryValues({
            'initial_altitude': (alt_i_climb, 'm'),
            'final_altitude': (alt_f_climb, 'm'),
            'initial_mach': (mach_i_climb, 'unitless'),
            'final_mach': (mach_f_climb, 'unitless'),
            'fix_initial': (False, 'unitless'),
            'input_initial': (True, 'unitless'),
            'use_polynomial_control': (False, 'unitless'),
        }),
        core_subsystems=default_mission_subsystems,
        subsystem_options={'core_aerodynamics': {'method': 'computed'}},
        transcription=transcription_climb,
    )

    cruise_options = EnergyPhase(
        'test_cruise',
        user_options=AviaryValues({
            'initial_altitude': (alt_min_cruise, 'm'),
            'final_altitude': (alt_max_cruise, 'm'),
            'initial_mach': (cruise_mach, 'unitless'),
            'final_mach': (cruise_mach, 'unitless'),
            'required_available_climb_rate': (300, 'ft/min'),
            'fix_initial': (False, 'unitless'),
        }),
        core_subsystems=default_mission_subsystems,
        subsystem_options={'core_aerodynamics': {'method': 'computed'}},
        transcription=transcription_cruise,
    )

    descent_options = EnergyPhase(
        'test_descent',
        user_options=AviaryValues({
            'final_altitude': (alt_f_descent, 'm'),
            'initial_altitude': (alt_i_descent, 'm'),
            'initial_mach': (mach_i_descent, 'unitless'),
            'final_mach': (mach_f_descent, 'unitless'),
            'fix_initial': (False, 'unitless'),
            'use_polynomial_control': (False, 'unitless'),
        }),
        core_subsystems=default_mission_subsystems,
        subsystem_options={'core_aerodynamics': {'method': 'computed'}},
        transcription=transcription_descent,
    )

    landing_options = Landing(
        ref_wing_area=aviary_inputs.get_val(Aircraft.Wing.AREA, units='ft**2'),
        Cl_max_ldg=aviary_inputs.get_val(
            Mission.Landing.LIFT_COEFFICIENT_MAX)  # no units
    )

    preprocess_crewpayload(aviary_inputs)

    # Upstream static analysis for aero
    prob.model.add_subsystem(
        'pre_mission',
        CorePreMission(aviary_options=aviary_inputs,
                       subsystems=default_premission_subsystems),
        promotes_inputs=['aircraft:*', 'mission:*'],
        promotes_outputs=['aircraft:*', 'mission:*'])

    # directly connect phases (strong_couple = True), or use linkage constraints (weak
    # coupling / strong_couple=False)
    strong_couple = False

    takeoff = takeoff_options.build_phase(False)

    climb = climb_options.build_phase(aviary_options=aviary_inputs)

    cruise = cruise_options.build_phase(aviary_options=aviary_inputs)

    descent = descent_options.build_phase(aviary_options=aviary_inputs)

    landing = landing_options.build_phase(False)

    prob.model.add_subsystem(
        'takeoff', takeoff, promotes_inputs=['aircraft:*', 'mission:*'],
        promotes_outputs=['mission:*'])

    traj = prob.model.add_subsystem('traj', dm.Trajectory())

    # if fix_initial is false, can we always set input_initial to be true for
    # necessary states, and then ignore if we use a linkage?
    climb.set_time_options(fix_initial=True, fix_duration=False, units='s',
                           duration_bounds=(t_duration_climb*0.5, t_duration_climb*2),
                           duration_ref=t_duration_climb)
    cruise.set_time_options(fix_initial=False, fix_duration=False, units='s',
                            duration_bounds=(t_duration_cruise*0.5, t_duration_cruise*2),
                            duration_ref=t_duration_cruise,
                            initial_bounds=(t_duration_climb*0.5, t_duration_climb*2))
    descent.set_time_options(
        fix_initial=False, fix_duration=False, units='s',
        duration_bounds=(t_duration_descent*0.5, t_duration_descent*2),
        duration_ref=t_duration_descent,
        initial_bounds=(
            (t_duration_cruise + t_duration_climb)*0.5,
            (t_duration_cruise + t_duration_climb)*2))

    traj.add_phase('climb', climb)

    traj.add_phase('cruise', cruise)

    traj.add_phase('descent', descent)

    prob.model.add_subsystem(
        'landing', landing, promotes_inputs=['aircraft:*', 'mission:*'],
        promotes_outputs=['mission:*'])

    traj.link_phases(["climb", "cruise", "descent"], [
                     "time", Dynamic.Mission.MASS, Dynamic.Mission.DISTANCE], connected=strong_couple)

    traj = setup_trajectory_params(prob.model, traj, aviary_inputs)

    ##################################
    # Connect in Takeoff and Landing #
    ##################################
    prob.model.connect(Mission.Takeoff.FINAL_MASS,
                       'traj.climb.initial_states:mass')
    prob.model.connect(Mission.Takeoff.GROUND_DISTANCE,
                       'traj.climb.initial_states:distance')

    prob.model.connect('traj.descent.states:mass',
                       Mission.Landing.TOUCHDOWN_MASS, src_indices=[-1])
    # TODO: approach velocity should likely be connected
    prob.model.connect('traj.descent.control_values:altitude', Mission.Landing.INITIAL_ALTITUDE,
                       src_indices=[-1])

    ##########################
    # Constraints            #
    ##########################

    ecomp = om.ExecComp('fuel_burned = initial_mass - descent_mass_final',
                        initial_mass={'units': 'lbm', 'shape': 1},
                        descent_mass_final={'units': 'lbm', 'shape': 1},
                        fuel_burned={'units': 'lbm', 'shape': 1})

    prob.model.add_subsystem('fuel_burn', ecomp,
                             promotes_inputs=[
                                 ('initial_mass', Mission.Design.GROSS_MASS)],
                             promotes_outputs=['fuel_burned'])

    prob.model.connect("traj.descent.states:mass",
                       "fuel_burn.descent_mass_final", src_indices=[-1])

    # TODO: need to add some sort of check that this value is less than the fuel capacity
    # TODO: need to update this with actual FLOPS value, this gives unrealistic
    # appearance of accuracy
    # TODO: the overall_fuel variable is the burned fuel plus the reserve, but should
    # also include the unused fuel, and the hierarchy variable name should be more clear
    ecomp = om.ExecComp('overall_fuel = fuel_burned + fuel_reserve',
                        fuel_burned={'units': 'lbm', 'shape': 1},
                        fuel_reserve={'units': 'lbm', 'val': 2173.},
                        overall_fuel={'units': 'lbm'})
    prob.model.add_subsystem('fuel_calc', ecomp,
                             promotes_inputs=['fuel_burned'],
                             promotes_outputs=['overall_fuel'])

    ecomp = om.ExecComp(
        'mass_resid = operating_empty_mass + overall_fuel + payload_mass '
        '- initial_mass',
        operating_empty_mass={'units': 'lbm'},
        overall_fuel={'units': 'lbm'},
        payload_mass={'units': 'lbm'},
        initial_mass={'units': 'lbm'},
        mass_resid={'units': 'lbm'})

    prob.model.add_subsystem(
        'mass_constraint', ecomp,
        promotes_inputs=[
            ('operating_empty_mass', Aircraft.Design.OPERATING_MASS),
            'overall_fuel',
            ('payload_mass', Aircraft.CrewPayload.TOTAL_PAYLOAD_MASS),
            ('initial_mass', Mission.Design.GROSS_MASS)],
        promotes_outputs=['mass_resid'])

    prob.model.add_constraint('mass_resid', equals=0.0, ref=1.0)

    ##########################
    # Add Objective Function #
    ##########################

    # This is an example of a overall mission objective
    # create a compound objective that minimizes climb time and maximizes final mass
    # we are maxing final mass b/c we don't have an independent value for fuel_mass yet
    # we are going to normalize these (making each of the sub-objectives approx = 1 )
    # TODO: change the scaling on climb_duration
    prob.model.add_subsystem(
        "regularization",
        om.ExecComp(
            "reg_objective = fuel_mass/1500",
            reg_objective=0.0,
            fuel_mass={"units": "lbm", "shape": 1},
        ),
        promotes_outputs=['reg_objective']
    )
    # connect the final mass from cruise into the objective
    prob.model.connect(Mission.Design.FUEL_MASS, "regularization.fuel_mass")

    prob.model.add_objective('reg_objective', ref=1)

    # Set initial default values for all LEAPS aircraft variables.
    set_aviary_initial_values(prob.model, aviary_inputs)

    prob.model.add_subsystem(
        'input_sink',
        VariablesIn(aviary_options=aviary_inputs),
        promotes_inputs=['*'],
        promotes_outputs=['*']
    )

    prob.setup(force_alloc_complex=True)

    ###########################################
    # Intial Settings for States and Controls #
    ###########################################

    prob.set_val('traj.climb.t_initial', t_i_climb, units='s')
    prob.set_val('traj.climb.t_duration', t_duration_climb, units='s')

    prob.set_val('traj.climb.controls:altitude', climb.interp(
        Dynamic.Mission.ALTITUDE, ys=[alt_i_climb, alt_f_climb]), units='m')
    prob.set_val(
        'traj.climb.controls:mach', climb.interp(
            Dynamic.Mission.MACH, ys=[mach_i_climb, mach_f_climb]), units='unitless')
    prob.set_val('traj.climb.states:mass', climb.interp(
        Dynamic.Mission.MASS, ys=[mass_i_climb, mass_f_climb]), units='kg')
    prob.set_val('traj.climb.states:distance', climb.interp(
        Dynamic.Mission.DISTANCE, ys=[distance_i_climb, distance_f_climb]), units='m')

    prob.set_val('traj.cruise.t_initial', t_i_cruise, units='s')
    prob.set_val('traj.cruise.t_duration', t_duration_cruise, units='s')

    prob.set_val('traj.cruise.polynomial_controls:altitude', cruise.interp(
        Dynamic.Mission.ALTITUDE, ys=[alt_i_cruise, alt_f_cruise]), units='m')
    prob.set_val(
        'traj.cruise.polynomial_controls:mach', cruise.interp(
            Dynamic.Mission.MACH, ys=[cruise_mach, cruise_mach]), units='unitless')
    prob.set_val('traj.cruise.states:mass', cruise.interp(
        Dynamic.Mission.MASS, ys=[mass_i_cruise, mass_f_cruise]), units='kg')
    prob.set_val('traj.cruise.states:distance', cruise.interp(
        Dynamic.Mission.DISTANCE, ys=[distance_i_cruise, distance_f_cruise]), units='m')

    prob.set_val('traj.descent.t_initial', t_i_descent, units='s')
    prob.set_val('traj.descent.t_duration', t_duration_descent, units='s')

    prob.set_val('traj.descent.controls:altitude', descent.interp(
        Dynamic.Mission.ALTITUDE, ys=[alt_i_descent, alt_f_descent]), units='m')
    prob.set_val(
        'traj.descent.controls:mach', descent.interp(
            Dynamic.Mission.MACH, ys=[mach_i_descent, mach_f_descent]), units='unitless')
    prob.set_val('traj.descent.states:mass', descent.interp(
        Dynamic.Mission.MASS, ys=[mass_i_descent, mass_f_descent]), units='kg')
    prob.set_val('traj.descent.states:distance', descent.interp(
        Dynamic.Mission.DISTANCE, ys=[distance_i_descent, distance_f_descent]), units='m')

    # Turn off solver printing so that the SNOPT output is readable.
    prob.set_solver_print(level=0)

    dm.run_problem(prob, simulate=sim, make_plots=False, simulate_kwargs={
                   'times_per_seg': 100, 'atol': 1e-9, 'rtol': 1e-9},
                   solution_record_file='N3CC_sizing.db')
    # prob.run_model()
    # z=prob.check_totals(method='cs', step=2e-40, compact_print=False)
    # exit()
    prob.record("final")
    prob.cleanup()

    return prob


@use_tempdirs
class ProblemPhaseTestCase(unittest.TestCase):

    @require_pyoptsparse(optimizer="SNOPT")
    def bench_test_sizing_N3CC(self):

        prob = run_trajectory(sim=False)

        times_climb = prob.get_val('traj.climb.timeseries.time', units='s')
        altitudes_climb = prob.get_val(
            'traj.climb.timeseries.altitude', units='m')
        masses_climb = prob.get_val('traj.climb.timeseries.mass', units='kg')
        distances_climb = prob.get_val('traj.climb.timeseries.distance', units='m')
        velocities_climb = prob.get_val(
            'traj.climb.timeseries.velocity', units='m/s')
        thrusts_climb = prob.get_val('traj.climb.timeseries.thrust_net_total', units='N')
        times_cruise = prob.get_val('traj.cruise.timeseries.time', units='s')
        altitudes_cruise = prob.get_val(
            'traj.cruise.timeseries.altitude', units='m')
        masses_cruise = prob.get_val('traj.cruise.timeseries.mass', units='kg')
        distances_cruise = prob.get_val('traj.cruise.timeseries.distance', units='m')
        velocities_cruise = prob.get_val(
            'traj.cruise.timeseries.velocity', units='m/s')
        thrusts_cruise = prob.get_val(
            'traj.cruise.timeseries.thrust_net_total', units='N')
        times_descent = prob.get_val('traj.descent.timeseries.time', units='s')
        altitudes_descent = prob.get_val(
            'traj.descent.timeseries.altitude', units='m')
        masses_descent = prob.get_val('traj.descent.timeseries.mass', units='kg')
        distances_descent = prob.get_val('traj.descent.timeseries.distance', units='m')
        velocities_descent = prob.get_val(
            'traj.descent.timeseries.velocity', units='m/s')
        thrusts_descent = prob.get_val(
            'traj.descent.timeseries.thrust_net_total', units='N')

        expected_times_s_climb = [[120.], [163.76325828], [224.14762249], [243.2590673],
                                  [243.2590673], [336.41088085], [464.94137099],
                                  [505.62083166], [505.62083166], [626.47618404],
                                  [793.2319034], [846.00951843], [846.00951843],
                                  [966.8648708], [1133.62059016], [1186.39820519],
                                  [1186.39820519], [1279.55001873], [1408.08050887],
                                  [1448.75996955], [1448.75996955], [1492.52322782],
                                  [1552.90759204], [1572.01903685]]
        expected_altitudes_m_climb = [[10.668], [0.], [835.43072799], [1237.19027083],
                                      [1237.19027083], [3154.43458807], [5478.53141388],
                                      [6035.03506592], [6035.03506592], [7369.93558792],
                                      [8611.93242084], [8931.907682], [8931.907682],
                                      [9555.56028912], [10196.52547754], [10341.80261401],
                                      [10341.80261401], [10526.85288051], [10653.9273369],
                                      [10668.], [10668.], [10668.], [10664.89939949], [10668.]]
        expected_masses_kg_climb = [[58331.64520977], [58289.14326023], [58210.1552262],
                                    [58183.99841695], [58183.99841695], [58064.21371439],
                                    [57928.02594919], [57893.85802859], [57893.85802859],
                                    [57803.5753272], [57701.17154376], [57671.70469428],
                                    [57671.70469428], [57609.99954455], [57532.67046087],
                                    [57509.28816794], [57509.28816794], [57468.36184014],
                                    [57411.39726843], [57392.97463799], [57392.97463799],
                                    [57372.80054331], [57344.30023996], [57335.11578186]]
        expected_distances_m_climb = [[1459.66454213], [6138.11750563], [15269.84014039],
                                      [18334.78691191], [18334.78691191], [34224.35398213],
                                      [57056.15756331], [64259.82264908], [64259.82264908],
                                      [85983.44219477], [116781.82372802], [
                                          126787.42506431],
                                      [126787.42506431], [149770.216723], [181655.376959],
                                      [191785.63505195], [191785.63505195], [
                                          210054.32022093],
                                      [236459.08240245], [245218.76016711], [
                                          245218.76016711],
                                      [254896.51199279], [268679.73823007], [273140.04867091]]
        expected_velocities_ms_climb = [[77.19291754], [132.35228283], [162.28279625],
                                        [166.11250634], [166.11250634], [176.30796789],
                                        [178.55049791], [178.67862048], [178.67862048],
                                        [181.75616771], [188.37687907], [189.78671051],
                                        [189.78671051], [190.67789763], [191.79702196],
                                        [193.51306342], [193.51306342], [199.28763405],
                                        [212.62699415], [218.00379223], [218.00379223],
                                        [224.21350812], [232.18822869], [234.25795132]]
        expected_thrusts_N_climb = [[82387.90445676], [105138.56964482], [113204.81413537],
                                    [110631.69881503], [110631.69881503], [97595.73505557],
                                    [74672.35033284], [68787.54639919], [68787.54639919],
                                    [56537.02658008], [46486.74146611], [43988.99250831],
                                    [43988.99250831], [39814.93772493], [36454.43040965],
                                    [35877.85248136], [35877.85248136], [35222.20684522],
                                    [34829.78305946], [34857.84665827], [34857.84665827],
                                    [34988.40432301], [35243.28462529], [35286.14075168]]

        expected_times_s_cruise = [[1572.01903685], [10224.85688577],
                                   [22164.00704809], [25942.70725365]]
        expected_altitudes_m_cruise = [[10668.], [10668.],
                                       [10668.], [10668.]]
        expected_masses_kg_cruise = [[57335.11578186], [53895.3524649],
                                     [49306.34176818], [47887.72131688]]
        expected_distances_m_cruise = [[273140.04867091], [2300136.11626779],
                                       [5096976.9738027], [5982167.54261146]]
        expected_velocities_ms_cruise = [[234.25795132], [234.25795132],
                                         [234.25795132], [234.25795132]]
        expected_thrusts_N_cruise = [[28998.46944214], [28027.44677784],
                                     [26853.54343662], [26522.10071819]]

        expected_times_s_descent = [[25942.70725365], [26017.49838255], [26120.69487534],
                                    [26153.35621248], [26153.35621248], [26306.10095576],
                                    [26516.85752584], [26583.56125968], [26583.56125968],
                                    [26765.15964678], [27015.72835621], [27095.03249603],
                                    [27095.03249603], [27247.77723931], [27458.5338094],
                                    [27525.23754324], [27525.23754324], [27600.02867214],
                                    [27703.22516493], [27735.88650207]]
        expected_altitudes_m_descent = [[10668.], [10668.], [10065.85989164],
                                        [9813.33438203], [9813.33438203], [8725.84180063],
                                        [7399.4358998], [6995.36734403], [6995.36734403],
                                        [5880.98032937], [4300.0193441], [3787.54388607],
                                        [3787.54388607], [2768.24427675], [1328.48487414],
                                        [872.88952055], [872.88952055], [393.61709386],
                                        [10.668], [10.668]]
        expected_masses_kg_descent = [[47887.72131688], [47887.72131688], [47887.72131688],
                                      [47887.72131688], [47887.72131688], [47887.72131688],
                                      [47887.72131688], [47887.72131688], [47887.72131688],
                                      [47887.87994829], [47886.14050369], [47884.40804261],
                                      [47884.40804261], [47872.68009732], [47849.34258173],
                                      [47842.09391697], [47842.09391697], [47833.150133],
                                      [47820.60083267], [47816.69389115]]
        expected_distances_m_descent = [[5982167.54261146], [5998182.9986382], [6017437.15474761],
                                        [6023441.80485498], [6023441.80485498], [
            6050450.80904601],
            [6085202.09693457], [6095472.55993089], [
            6095472.55993089],
            [6122427.64793619], [6158176.53029461], [
            6169433.26612421],
            [6169433.26612421], [6191073.79296936], [
            6220839.65410345],
            [6230195.89769052], [6230195.89769052], [
            6240573.15704622],
            [6253740.15118502], [6257352.4]]
        expected_velocities_ms_descent = [[234.25795132], [197.64415171], [182.5029101],
                                          [181.15994177], [181.15994177], [172.42254637],
                                          [156.92424445], [152.68023428], [152.68023428],
                                          [145.15327267], [141.83129659], [141.72294853],
                                          [141.72294853], [141.82174008], [140.7384589],
                                          [139.65952688], [139.65952688], [136.46721905],
                                          [115.68627038], [102.07377559]]
        expected_thrusts_N_descent = [[0.], [8.11038373e-13], [4.89024663e-13],
                                      [-3.80747087e-14], [0.], [1.10831527e-12],
                                      [7.77983996e-13], [-6.64041439e-16], [0.],
                                      [0.], [-5.7108371e-14], [-8.28280737e-14],
                                      [0.], [1.14382967e-13], [5.36023712e-14],
                                      [-3.04972888e-14], [0.], [-1.0682523e-13],
                                      [-6.30544276e-14], [5.37050667e-15]]
        expected_times_s_climb = np.array(expected_times_s_climb)
        expected_altitudes_m_climb = np.array(expected_altitudes_m_climb)
        expected_masses_kg_climb = np.array(expected_masses_kg_climb)
        expected_distances_m_climb = np.array(expected_distances_m_climb)
        expected_velocities_ms_climb = np.array(expected_velocities_ms_climb)
        expected_thrusts_N_climb = np.array(expected_thrusts_N_climb)

        expected_times_s_cruise = np.array(expected_times_s_cruise)
        expected_altitudes_m_cruise = np.array(expected_altitudes_m_cruise)
        expected_masses_kg_cruise = np.array(expected_masses_kg_cruise)
        expected_distances_m_cruise = np.array(expected_distances_m_cruise)
        expected_velocities_ms_cruise = np.array(expected_velocities_ms_cruise)
        expected_thrusts_N_cruise = np.array(expected_thrusts_N_cruise)

        expected_times_s_descent = np.array(expected_times_s_descent)
        expected_altitudes_m_descent = np.array(expected_altitudes_m_descent)
        expected_masses_kg_descent = np.array(expected_masses_kg_descent)
        expected_distances_m_descent = np.array(expected_distances_m_descent)
        expected_velocities_ms_descent = np.array(expected_velocities_ms_descent)
        expected_thrusts_N_descent = np.array(expected_thrusts_N_descent)

        # Check Objective and other key variables to a reasonably tight tolerance.

        # TODO: update truth values once everyone is using latest Dymos
        rtol = 5e-2

        # Mass at the end of Descent
        assert_near_equal(masses_descent[-1],
                          expected_masses_kg_descent[-1], tolerance=rtol)

        # Range at the end of Descent
        assert_near_equal(distances_descent[-1],
                          expected_distances_m_descent[-1], tolerance=rtol)

        # Flight time
        assert_near_equal(times_descent[-1],
                          expected_times_s_descent[-1], tolerance=rtol)

        # Check mission values.

        # NOTE rtol = 0.05 = 5% different from truth (first timeseries)
        #      atol = 2 = no more than +/-2 meter/second/kg difference between values
        #      atol_altitude - 30 ft. There is occasional time-shifting with the N3CC
        #                      model during climb and descent so we need a looser
        #                      absolute tolerance for the points near the ground.
        rtol = .05
        atol = 2.0
        atol_altitude = 30.0

        # FLIGHT PATH
        # CLIMB
        warn_timeseries_near_equal(
            times_climb, altitudes_climb, expected_times_s_climb,
            expected_altitudes_m_climb, abs_tolerance=atol_altitude, rel_tolerance=rtol)
        warn_timeseries_near_equal(
            times_climb, masses_climb, expected_times_s_climb,
            expected_masses_kg_climb, abs_tolerance=atol, rel_tolerance=rtol)
        warn_timeseries_near_equal(
            times_climb, distances_climb, expected_times_s_climb,
            expected_distances_m_climb, abs_tolerance=atol, rel_tolerance=rtol)
        warn_timeseries_near_equal(
            times_climb, velocities_climb, expected_times_s_climb,
            expected_velocities_ms_climb, abs_tolerance=atol, rel_tolerance=rtol)
        warn_timeseries_near_equal(
            times_climb, thrusts_climb, expected_times_s_climb,
            expected_thrusts_N_climb, abs_tolerance=atol, rel_tolerance=rtol)

        # CRUISE
        warn_timeseries_near_equal(
            times_cruise, altitudes_cruise, expected_times_s_cruise,
            expected_altitudes_m_cruise, abs_tolerance=atol, rel_tolerance=rtol)
        warn_timeseries_near_equal(
            times_cruise, masses_cruise, expected_times_s_cruise,
            expected_masses_kg_cruise, abs_tolerance=atol, rel_tolerance=rtol)
        warn_timeseries_near_equal(
            times_cruise, distances_cruise, expected_times_s_cruise,
            expected_distances_m_cruise, abs_tolerance=atol, rel_tolerance=rtol)
        warn_timeseries_near_equal(
            times_cruise, velocities_cruise, expected_times_s_cruise,
            expected_velocities_ms_cruise, abs_tolerance=atol, rel_tolerance=rtol)
        warn_timeseries_near_equal(
            times_cruise, thrusts_cruise, expected_times_s_cruise,
            expected_thrusts_N_cruise, abs_tolerance=atol, rel_tolerance=rtol)

        # DESCENT
        warn_timeseries_near_equal(
            times_descent, altitudes_descent, expected_times_s_descent,
            expected_altitudes_m_descent, abs_tolerance=atol_altitude,
            rel_tolerance=rtol)
        warn_timeseries_near_equal(
            times_descent, masses_descent, expected_times_s_descent,
            expected_masses_kg_descent, abs_tolerance=atol, rel_tolerance=rtol)
        warn_timeseries_near_equal(
            times_descent, distances_descent, expected_times_s_descent,
            expected_distances_m_descent, abs_tolerance=atol, rel_tolerance=rtol)
        warn_timeseries_near_equal(
            times_descent, velocities_descent, expected_times_s_descent,
            expected_velocities_ms_descent, abs_tolerance=atol, rel_tolerance=rtol)
        warn_timeseries_near_equal(
            times_descent, thrusts_descent, expected_times_s_descent,
            expected_thrusts_N_descent, abs_tolerance=atol, rel_tolerance=rtol)


if __name__ == '__main__':
    z = ProblemPhaseTestCase()
    z.bench_test_sizing_N3CC()



import numpy as np
import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.enums import GASPEngineType
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


def sig(x):
    return 1 / (1 + np.exp(-100 * x))


def dsig(x):
    return 100 * np.exp(-100 * x) / (np.exp(-100 * x) + 1) ** 2


class EquipAndUsefulLoadMass(om.ExplicitComponent):
    def initialize(self):

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        add_aviary_input(self, Aircraft.Design.EQUIPMENT_MASS_COEFFICIENTS, val=[
                         1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13], units="unitless")
        add_aviary_input(self, Mission.Design.GROSS_MASS, val=175400)
        add_aviary_input(self, Aircraft.Fuselage.LENGTH, val=128)
        add_aviary_input(self, Aircraft.Wing.SPAN, val=117.8)
        add_aviary_input(self, Aircraft.LandingGear.TOTAL_MASS, val=200)
        add_aviary_input(self, Aircraft.Controls.TOTAL_MASS, val=150)
        add_aviary_input(self, Aircraft.Wing.AREA, val=150)
        add_aviary_input(self, Aircraft.HorizontalTail.AREA, val=150)
        add_aviary_input(self, Aircraft.VerticalTail.AREA, val=150)
        add_aviary_input(self, Aircraft.Fuselage.PRESSURE_DIFFERENTIAL, val=7.5)
        add_aviary_input(self, Aircraft.Fuselage.AVG_DIAMETER, val=13.1)
        add_aviary_input(self, Aircraft.Engine.SCALED_SLS_THRUST, val=28690)
        add_aviary_input(self, Aircraft.Fuel.WING_FUEL_FRACTION, val=0.5)
        add_aviary_input(self, Aircraft.Design.EXTERNAL_SUBSYSTEMS_MASS, val=0.)

        add_aviary_output(self, Aircraft.Design.FIXED_USEFUL_LOAD, val=0)
        add_aviary_output(self, Aircraft.Design.FIXED_EQUIPMENT_MASS, val=0)

        self.declare_partials(Aircraft.Design.FIXED_USEFUL_LOAD, '*', val=0.0)
        self.declare_partials(Aircraft.Design.FIXED_EQUIPMENT_MASS, '*', val=0.0)
        self.declare_partials(Aircraft.Design.FIXED_EQUIPMENT_MASS,
                              Aircraft.Design.EXTERNAL_SUBSYSTEMS_MASS, val=1./GRAV_ENGLISH_LBM)

    def compute(self, inputs, outputs):

        options: AviaryValues = self.options["aviary_options"]
        PAX = options.get_val(Aircraft.CrewPayload.NUM_PASSENGERS, units='unitless')
        smooth = options.get_val(
            Aircraft.Design.SMOOTH_MASS_DISCONTINUITIES, units='unitless')
        mass_coeff = inputs[Aircraft.Design.EQUIPMENT_MASS_COEFFICIENTS]
        gross_wt_initial = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        num_engines = self.options['aviary_options'].get_val(
            Aircraft.Propulsion.TOTAL_NUM_ENGINES, units='unitless')
        fus_len = inputs[Aircraft.Fuselage.LENGTH]
        wingspan = inputs[Aircraft.Wing.SPAN]
        if options.get_val(Aircraft.LandingGear.FIXED_GEAR, units='unitless'):
            gear_type = 1
        else:
            gear_type = 0

        landing_gear_wt = inputs[Aircraft.LandingGear.TOTAL_MASS] * \
            GRAV_ENGLISH_LBM
        control_wt = inputs[Aircraft.Controls.TOTAL_MASS] * GRAV_ENGLISH_LBM
        wing_area = inputs[Aircraft.Wing.AREA]
        htail_area = inputs[Aircraft.HorizontalTail.AREA]
        vtail_area = inputs[Aircraft.VerticalTail.AREA]
        p_diff_fus = inputs[Aircraft.Fuselage.PRESSURE_DIFFERENTIAL]
        cabin_width = inputs[Aircraft.Fuselage.AVG_DIAMETER]
        Fn_SLS = inputs[Aircraft.Engine.SCALED_SLS_THRUST]
        fuel_vol_frac = inputs[Aircraft.Fuel.WING_FUEL_FRACTION]
        subsystems_wt = inputs[Aircraft.Design.EXTERNAL_SUBSYSTEMS_MASS]

        engine_type = options.get_val(Aircraft.Engine.TYPE, units='unitless')[0]

        APU_wt = 0.0
        if PAX > 35.0:
            APU_wt = 26.2 * PAX**0.944 - 13.6 * PAX
        if ~(
            -1e-5 < mass_coeff[0] < 1e-5
        ):  # note: this technically creates a discontinuity
            APU_wt = mass_coeff[0] * GRAV_ENGLISH_LBM

        num_pilots = 1.0
        if PAX > 9.0:
            num_pilots = 2.0
        if engine_type is GASPEngineType.TURBOJET and PAX > 5.0:
            num_pilots = 2.0
        if PAX >= 251.0:
            num_pilots = 3.0

        instrument_wt = (
            mass_coeff[1]
            * gross_wt_initial**0.386
            * num_engines**0.687
            * num_pilots**0.31
            * fus_len**0.05
            * wingspan**0.696
        )
        gear_val = 1 - gear_type
        hydraulic_wt = (
            mass_coeff[2] * control_wt + mass_coeff[3] * landing_gear_wt * gear_val
        )

        electrical_wt = 16.0 * PAX + 170.0
        if PAX <= 12.0:
            electrical_wt = 0.03217 * gross_wt_initial - 20.0
        if num_engines == 1.0:
            electrical_wt = 0.00778 * gross_wt_initial + 33.0

        avionics_wt = 27.0

        if smooth:

            avionics_wt = 35.538 * np.exp(0.0002 * gross_wt_initial)

        else:
            if (
                gross_wt_initial >= 3000.0
            ):  # note: this technically creates a discontinuity
                avionics_wt = 65.0
            if (
                gross_wt_initial >= 5500.0
            ):  # note: this technically creates a discontinuity
                avionics_wt = 113.0
            if (
                gross_wt_initial >= 7500.0
            ):  # note: this technically creates a discontinuity
                avionics_wt = 163.0
            if (
                gross_wt_initial >= 11000.0
            ):  # note: this technically creates a discontinuity
                avionics_wt = 340.0

        if PAX >= 20.0 and PAX < 30.0:
            avionics_wt = 400.0
        elif PAX >= 30.0 and PAX <= 50.0:
            avionics_wt = 500.0
        elif PAX > 50.0:
            avionics_wt = 600.0
        if PAX > 100.0:
            avionics_wt = 2.8 * PAX + 1010.0
        if ~(
            -1e-5 < mass_coeff[4] < 1e-5
        ):  # note: this technically creates a discontinuity !WILL NOT CHANGE
            avionics_wt = mass_coeff[4] * GRAV_ENGLISH_LBM

        air_conditioning_wt = 5.0

        if gross_wt_initial > 3500.0:  # note: this technically creates a discontinuity
            air_conditioning_wt = (
                mass_coeff[5]
                * (1.5 + p_diff_fus)
                * (0.358 * fus_len * cabin_width**2) ** 0.5
            )

        SSUM = wing_area + htail_area + vtail_area
        icing_wt = 22.7 * (SSUM**0.5) - 385.0

        if smooth:
            pass
        else:
            if icing_wt < 0.0:  # note: this technically creates a discontinuity
                icing_wt = 0.0
        if ~(
            -1e-5 < mass_coeff[6] < 1e-5
        ):  # note: this technically creates a discontinuity !WILL NOT CHANGE
            icing_wt = mass_coeff[6] * GRAV_ENGLISH_LBM

        aux_wt = 0.0

        if smooth:
            aux_wt = 3 * sig((gross_wt_initial - 3000) / 3000)

        else:
            if (
                gross_wt_initial > 3000.0
            ):  # note: this technically creates a discontinuity
                aux_wt = 3.0

        if PAX >= 9.0:
            aux_wt = 10.0
        if PAX > 19.0:
            aux_wt = 20.0
        if PAX > 74.0:
            aux_wt = 50.0

        CPX = 28.0 + 10.516 * (cabin_width - 5.667)

        if smooth:
            CPX = (
                28 * sig((28 - CPX) / 28)
                + CPX * sig((CPX - 28) / 28) * sig((62 - CPX) / 62)
                + 62 * sig((CPX - 62) / 62)
            )

        else:
            if cabin_width <= 5.667:  # note: this technically creates a discontinuity
                CPX = 28.0
            if cabin_width > 8.90:  # note: this technically creates a discontinuity
                CPX = 62.0

        furnishing_wt = CPX * PAX + 310.0
        if PAX > 80:
            furnishing_wt = 118.4 * PAX - 4190.0
        if (
            gross_wt_initial <= 10000.0
        ):  # note: this technically creates a discontinuity #TODO: Doesn't occur in large single aisle
            furnishing_wt = 0.065 * gross_wt_initial - 59.0

        if smooth:
            pass
        else:
            if furnishing_wt <= 30.0:  # note: this technically creates a discontinuity
                furnishing_wt = 30.0
        if ~(
            -1e-5 < mass_coeff[7] < 1e-5
        ):  # note: this technically creates a discontinuity #WONT CHANGE
            furnishing_wt = mass_coeff[7] * GRAV_ENGLISH_LBM
        fixed_equip_wt = (
            APU_wt
            + instrument_wt
            + hydraulic_wt
            + electrical_wt
            + avionics_wt
            + air_conditioning_wt
            + icing_wt
            + aux_wt
            + furnishing_wt
            + subsystems_wt
        )

        outputs[Aircraft.Design.FIXED_EQUIPMENT_MASS] = fixed_equip_wt / \
            GRAV_ENGLISH_LBM

        num_flight_attendants = 0.0
        if PAX >= 20.0:
            num_flight_attendants = 1.0
        if PAX >= 51.0:
            num_flight_attendants = 2.0
        if PAX >= 101.0:
            num_flight_attendants = 3.0
        if PAX >= 151.0:
            num_flight_attendants = 4.0
        if PAX >= 201.0:
            num_flight_attendants = 5.0
        if PAX >= 251.0:
            num_flight_attendants = 6.0

        # note: the average weight of a pilot was calulated using the following equation:
        # avg_wt = pct_male*avg_wt_male + pct_female*avg_wt_female where
        # pct_male = the percentage of US airline pilots that are male (based on data from
        # the center for aviation in 2018, which listed this percentage as 95.6%, and slightly
        # deflated to account for the upward trend in female pilots, resulting in an estimated
        # percentage of 94%)
        # avg_wt_male is the average weight of males according to the CDC, and is 199.8 lbf
        # pct_female is calculated from the same methods as pct_male, and results in 6%
        # avg_wt_female is the average weight of females according to the CDC, and is 170.8 lbf
        # the resulting value is that the average weight of the US airline pilot is 198 lbf
        pilot_wt = 198 * num_pilots
        # note: the average weight of a flight attendant was calulated using the following equation:
        # avg_wt = pct_male*avg_wt_male + pct_female*avg_wt_female where
        # pct_male = the percentage of US flight attendants that are male (based on data from
        # the women in aerospace international organization in 2020, which listed this percentage as
        # 20.8%)
        # avg_wt_male is the average weight of males according to the CDC, and is 199.8 lbf
        # pct_female is calculated from the same methods as pct_male, and results in 79.2%
        # avg_wt_female is the average weight of females according to the CDC, and is 170.8 lbf
        # the resulting value is that the average weight of the US flight attendant is 177 lbf
        flight_attendant_wt = 177 * num_flight_attendants

        crew_bag_wt = 25.0 * num_pilots
        if PAX >= 20.0:
            crew_bag_wt = 10.0 * (num_pilots + num_flight_attendants) + 25.0
        if PAX >= 40.0:
            crew_bag_wt = (
                20.0 * (num_flight_attendants + num_pilots) + 25.0 * num_pilots
            )

        if engine_type is GASPEngineType.TURBOJET:
            oil_per_eng_wt = 0.0054 * Fn_SLS + 12.0
        # elif engine_type is GASPEngineType.TURBOSHAFT or engine_type is GASPEngineType.TURBOPROP:
        #     oil_per_eng_wt = 0.0124 * Fn_SLS + 14
        # else:
        #     oil_per_eng_wt = 0.062 * (Fn_SLS - 100) + 11
        else:
            # Other engine types are currently not supported in Aviary
            oil_per_eng_wt = 0

        oil_wt = num_engines * oil_per_eng_wt

        lavatories = 0.0
        if PAX > 25.0:
            lavatories = 1.0
        if PAX >= 51.0:
            lavatories = 2.0
        if PAX >= 101.0:
            lavatories = 3.0
        if PAX >= 151.0:
            lavatories = 4.0
        if PAX >= 201.0:
            lavatories = 5.0
        if PAX >= 251.0:
            lavatories = 6.0

        service_wt = 0.0
        if PAX > 9.0:
            service_wt = mass_coeff[8] * PAX * \
                GRAV_ENGLISH_LBM + 16.0 * lavatories

        water_wt = 0.0
        if PAX > 19.0:
            water_wt = mass_coeff[9] * (PAX + num_pilots +
                                        num_flight_attendants) * GRAV_ENGLISH_LBM

        emergency_wt = 0.0
        if PAX > 5.0:
            emergency_wt = 10.0
        if PAX > 9.0:
            emergency_wt = 15.0
        if PAX >= 35.0:
            emergency_wt = 25.0 * num_flight_attendants + 15.0
        if ~(-1e-5 < mass_coeff[10] < 1e-5):
            emergency_wt = mass_coeff[10] * GRAV_ENGLISH_LBM

        catering_wt = 0.0
        if PAX > 19.0:
            catering_wt = mass_coeff[11] * PAX * GRAV_ENGLISH_LBM

        trapped_fuel_wt = mass_coeff[12] * (wing_area**0.5) * fuel_vol_frac / 0.430
        if (
            fuel_vol_frac <= 0.075
        ):  # note: this technically creates a discontinuity # won't change
            trapped_fuel_wt = mass_coeff[12] * 0.18 * (wing_area**0.5)

        useful_wt = (
            pilot_wt
            + flight_attendant_wt
            + crew_bag_wt
            + oil_wt
            + service_wt
            + water_wt
            + emergency_wt
            + catering_wt
            + trapped_fuel_wt
        )

        outputs[Aircraft.Design.FIXED_USEFUL_LOAD] = useful_wt / \
            GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, partials):
        options = self.options['aviary_options']
        PAX = options.get_val(Aircraft.CrewPayload.NUM_PASSENGERS, units='unitless')
        smooth = options.get_val(
            Aircraft.Design.SMOOTH_MASS_DISCONTINUITIES, units='unitless')
        mass_coeff = inputs[Aircraft.Design.EQUIPMENT_MASS_COEFFICIENTS]
        gross_wt_initial = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        num_engines = self.options['aviary_options'].get_val(
            Aircraft.Propulsion.TOTAL_NUM_ENGINES, units='unitless')
        fus_len = inputs[Aircraft.Fuselage.LENGTH]
        wingspan = inputs[Aircraft.Wing.SPAN]
        if options.get_val(Aircraft.LandingGear.FIXED_GEAR, units='unitless'):
            gear_type = 1
        else:
            gear_type = 0
        landing_gear_wt = inputs[Aircraft.LandingGear.TOTAL_MASS] * \
            GRAV_ENGLISH_LBM
        control_wt = inputs[Aircraft.Controls.TOTAL_MASS] * GRAV_ENGLISH_LBM
        wing_area = inputs[Aircraft.Wing.AREA]
        htail_area = inputs[Aircraft.HorizontalTail.AREA]
        vtail_area = inputs[Aircraft.VerticalTail.AREA]
        p_diff_fus = inputs[Aircraft.Fuselage.PRESSURE_DIFFERENTIAL]
        cabin_width = inputs[Aircraft.Fuselage.AVG_DIAMETER]
        fuel_vol_frac = inputs[Aircraft.Fuel.WING_FUEL_FRACTION]

        engine_type = options.get_val(Aircraft.Engine.TYPE, units='unitless')[0]

        dAPU_wt_dmass_coeff_0 = 0.0
        if ~(
            -1e-5 < mass_coeff[0] < 1e-5
        ):  # note: this technically creates a discontinuity
            dAPU_wt_dmass_coeff_0 = GRAV_ENGLISH_LBM

        num_pilots = 1.0
        if PAX > 9.0:
            num_pilots = 2.0
        if engine_type is GASPEngineType.TURBOJET and PAX > 5.0:
            num_pilots = 2.0
        if PAX >= 251.0:
            num_pilots = 3.0

        dinstrument_wt_dmass_coeff_1 = (
            gross_wt_initial**0.386
            * num_engines**0.687
            * num_pilots**0.31
            * fus_len**0.05
            * wingspan**0.696
        )
        dinstrument_wt_dgross_wt_initial = (
            0.386
            * mass_coeff[1]
            * gross_wt_initial ** (0.386 - 1)
            * num_engines**0.687
            * num_pilots**0.31
            * fus_len**0.05
            * wingspan**0.696
        )
        dinstrument_wt_dfus_len = (
            0.05
            * mass_coeff[1]
            * gross_wt_initial**0.386
            * num_engines**0.687
            * num_pilots**0.31
            * fus_len ** (0.05 - 1)
            * wingspan**0.696
        )
        dinstrument_wt_dwingspan = (
            0.696
            * mass_coeff[1]
            * gross_wt_initial**0.386
            * num_engines**0.687
            * num_pilots**0.31
            * fus_len**0.05
            * wingspan ** (0.696 - 1)
        )

        gear_val = 1 - gear_type

        dhydraulic_wt_dmass_coeff_2 = control_wt
        dhydraulic_wt_dcontrol_wt = mass_coeff[2]
        dhydraulic_wt_dmass_coeff_3 = landing_gear_wt * gear_val
        dhydraulic_wt_dlanding_gear_weight = mass_coeff[3] * gear_val

        delectrical_wt_dgross_wt_initial = 0.0
        if PAX <= 12.0:
            delectrical_wt_dgross_wt_initial = 0.03217
        if num_engines == 1.0:
            delectrical_wt_dgross_wt_initial = 0.0078

        davionics_wt_dmass_coeff_4 = 0.0

        if smooth:

            davionics_wt_dgross_wt_initial = 0.0071076 * np.exp(
                0.0002 * gross_wt_initial
            )

        else:
            if (
                gross_wt_initial >= 3000.0
            ):  # note: this technically creates a discontinuity
                davionics_wt_dgross_wt_initial = 0.0
            if (
                gross_wt_initial >= 5500.0
            ):  # note: this technically creates a discontinuity
                davionics_wt_dgross_wt_initial = 0.0
            if (
                gross_wt_initial >= 7500.0
            ):  # note: this technically creates a discontinuity
                davionics_wt_dgross_wt_initial = 0.0
            if (
                gross_wt_initial >= 11000.0
            ):  # note: this technically creates a discontinuity
                davionics_wt_dgross_wt_initial = 0.0

        if PAX >= 20.0 and PAX < 30.0:
            davionics_wt_dgross_wt_initial = 0.0
        elif PAX >= 30.0 and PAX <= 50.0:
            davionics_wt_dgross_wt_initial = 0.0
        elif PAX > 50.0:
            davionics_wt_dgross_wt_initial = 0.0
        if PAX > 100.0:
            davionics_wt_dgross_wt_initial = 0.0
        if ~(
            -1e-5 < mass_coeff[4] < 1e-5
        ):  # note: this technically creates a discontinuity !WILL NOT CHANGE
            davionics_wt_dgross_wt_initial = 0.0
            davionics_wt_dmass_coeff_4 = GRAV_ENGLISH_LBM

        dair_conditioning_wt_dmass_coeff_5 = 0.0
        dair_conditioning_wt_dp_diff_fus = 0.0
        dair_conditioning_wt_dfus_len = 0.0
        dair_conditioning_wt_dcabin_width = 0.0
        if gross_wt_initial > 3500.0:  # note: this technically creates a discontinuity
            dair_conditioning_wt_dmass_coeff_5 = (1.5 + p_diff_fus) * (
                0.358 * fus_len * cabin_width**2
            ) ** 0.5
            dair_conditioning_wt_dp_diff_fus = (
                mass_coeff[5] * (0.358 * fus_len * cabin_width**2) ** 0.5
            )
            dair_conditioning_wt_dfus_len = (
                0.5
                * mass_coeff[5]
                * (1.5 + p_diff_fus)
                * (0.358 * fus_len * cabin_width**2) ** -0.5
                * 0.358
                * cabin_width**2
            )
            dair_conditioning_wt_dcabin_width = (
                0.5
                * mass_coeff[5]
                * (1.5 + p_diff_fus)
                * (0.358 * fus_len * cabin_width**2) ** -0.5
                * 2
                * 0.358
                * fus_len
                * cabin_width
            )

        SSUM = wing_area + htail_area + vtail_area
        icing_wt = 22.7 * (SSUM**0.5) - 385.0

        dicing_weight_dwing_area = 0.5 * 22.7 * (SSUM**-0.5)
        dicing_weight_dhtail_area = 0.5 * 22.7 * (SSUM**-0.5)
        dicing_weight_dvtail_area = 0.5 * 22.7 * (SSUM**-0.5)
        dicing_weight_dmass_coeff_6 = 0.0

        if smooth:
            pass
        else:
            if icing_wt < 0.0:  # note: this technically creates a discontinuity
                icing_wt = 0.0
                dicing_weight_dwing_area = 0.0
                dicing_weight_dhtail_area = 0.0
                dicing_weight_dvtail_area = 0.0
                dicing_weight_dmass_coeff_6 = 0.0
        if ~(
            -1e-5 < mass_coeff[6] < 1e-5
        ):  # note: this technically creates a discontinuity !WILL NOT CHANGE
            icing_wt = mass_coeff[6] * GRAV_ENGLISH_LBM
            dicing_weight_dwing_area = 0.0
            dicing_weight_dhtail_area = 0.0
            dicing_weight_dvtail_area = 0.0
            dicing_weight_dmass_coeff_6 = GRAV_ENGLISH_LBM

        if smooth:
            d_aux_wt_dgross_wt_initial = (
                3 * dsig((gross_wt_initial - 3000) / 3000) * 1 / 3000
            )
        else:
            if (
                gross_wt_initial > 3000.0
            ):  # note: this technically creates a discontinuity
                d_aux_wt_dgross_wt_initial = 0.0

        if PAX >= 9.0:
            d_aux_wt_dgross_wt_initial = 0.0
        if PAX > 19.0:
            d_aux_wt_dgross_wt_initial = 0.0
        if PAX > 74.0:
            d_aux_wt_dgross_wt_initial = 0.0

        CPX = 28.0 + 10.516 * (cabin_width - 5.667)
        dCPX_dcabin_width = 10.516

        if smooth:
            CPX_1 = (
                28 * sig((28 - CPX) / 28)
                + CPX * sig((CPX - 28) / 28) * sig((62 - CPX) / 62)
                + 62 * sig((CPX - 62) / 62)
            )

            dCPX_dcabin_width = (
                28 * dsig((28 - CPX) / 28) * -dCPX_dcabin_width
                + (
                    dCPX_dcabin_width * sig((CPX - 28) / 28)
                    + CPX * dsig((CPX - 28) / 28) * dCPX_dcabin_width
                )
                * sig((62 - CPX) / 62)
                + CPX
                * sig((CPX - 28) / 28)
                * dsig((62 - CPX) / 62)
                * -dCPX_dcabin_width
                + 62 * dsig((CPX - 62) / 62) * dCPX_dcabin_width
            )

            CPX = CPX_1
            dfurnishing_wt_dgross_wt_initial = 0.0
            dfurnishing_wt_dgross_wt_initial = 0.0
            dfurnishing_wt_dmass_coeff_7 = 0.0
        else:
            if cabin_width <= 5.667:  # note: this technically creates a discontinuity
                CPX = 28.0
                dCPX_dcabin_width = 0.0
                dfurnishing_wt_dgross_wt_initial = 0.0
                dfurnishing_wt_dmass_coeff_7 = 0.0
            if cabin_width > 8.90:  # note: this technically creates a discontinuity
                CPX = 62.0
                dCPX_dcabin_width = 0.0
                dfurnishing_wt_dgross_wt_initial = 0.0
                dfurnishing_wt_dmass_coeff_7 = 0.0

        furnishing_wt = CPX * PAX + 310.0
        dfurnishing_wt_dgross_wt_initial = 0.0
        dfurnishing_wt_dmass_coeff_7 = 0.0
        dfurnishing_wt_dcabin_width = PAX * dCPX_dcabin_width
        if PAX > 80:
            furnishing_wt = 118.4 * PAX - 4190.0
            dfurnishing_wt_dgross_wt_initial = 0.0
            dfurnishing_wt_dcabin_width = 0.0
            dfurnishing_wt_dmass_coeff_7 = 0.0
        if (
            gross_wt_initial <= 10000.0
        ):  # note: this technically creates a discontinuity #TODO: Doesn't occur in large single aisle
            furnishing_wt = 0.065 * gross_wt_initial - 59.0
            dfurnishing_wt_dgross_wt_initial = 0.065
            dfurnishing_wt_dcabin_width = 0.0
            dfurnishing_wt_dmass_coeff_7 = 0.0

        if smooth:
            pass
        else:
            if furnishing_wt <= 30.0:  # note: this technically creates a discontinuity
                furnishing_wt = 30.0
                dfurnishing_wt_dgross_wt_initial = 0.0
                dfurnishing_wt_dcabin_width = 0.0
                dfurnishing_wt_dmass_coeff_7 = 0.0
        if ~(
            -1e-5 < mass_coeff[7] < 1e-5
        ):  # note: this technically creates a discontinuity #WONT CHANGE
            furnishing_wt = mass_coeff[7] * GRAV_ENGLISH_LBM
            dfurnishing_wt_dmass_coeff_7 = GRAV_ENGLISH_LBM
            dfurnishing_wt_dcabin_width = 0.0
            dfurnishing_wt_dgross_wt_initial = 0.0

        dfixed_equip_mass_dmass_coeff_0 = dAPU_wt_dmass_coeff_0 / GRAV_ENGLISH_LBM
        dfixed_equip_mass_dmass_coeff_1 = dinstrument_wt_dmass_coeff_1 / GRAV_ENGLISH_LBM
        dfixed_equip_wt_dgross_wt_initial = (
            dinstrument_wt_dgross_wt_initial
            + delectrical_wt_dgross_wt_initial
            + dfurnishing_wt_dgross_wt_initial
            + davionics_wt_dgross_wt_initial
            + d_aux_wt_dgross_wt_initial
            + dfurnishing_wt_dcabin_width
        )
        dfixed_equip_mass_dfus_len = (
            dinstrument_wt_dfus_len + dair_conditioning_wt_dfus_len
        ) / GRAV_ENGLISH_LBM
        dfixed_equip_mass_dwingspan = dinstrument_wt_dwingspan / GRAV_ENGLISH_LBM
        dfixed_equip_mass_dmass_coeff_2 = dhydraulic_wt_dmass_coeff_2 / GRAV_ENGLISH_LBM
        dfixed_equip_wt_dcontrol_wt = dhydraulic_wt_dcontrol_wt
        dfixed_equip_mass_dmass_coeff_3 = dhydraulic_wt_dmass_coeff_3 / GRAV_ENGLISH_LBM
        dfixed_equip_wt_dlanding_gear_weight = dhydraulic_wt_dlanding_gear_weight
        dfixed_equip_mass_dmass_coeff_4 = davionics_wt_dmass_coeff_4 / GRAV_ENGLISH_LBM

        dfixed_equip_mass_dmass_coeff_5 = dair_conditioning_wt_dmass_coeff_5 / GRAV_ENGLISH_LBM
        dfixed_equip_mass_dp_diff_fus = dair_conditioning_wt_dp_diff_fus / GRAV_ENGLISH_LBM
        dfixed_equip_mass_dcabin_width = dair_conditioning_wt_dcabin_width / GRAV_ENGLISH_LBM

        dfixed_equip_mass_dwing_area = dicing_weight_dwing_area / GRAV_ENGLISH_LBM
        dfixed_equip_mass_dhtail_area = dicing_weight_dhtail_area / \
            GRAV_ENGLISH_LBM
        dfixed_equip_mass_dvtail_area = dicing_weight_dvtail_area / \
            GRAV_ENGLISH_LBM
        dfixed_equip_mass_dmass_coeff_6 = dicing_weight_dmass_coeff_6 / GRAV_ENGLISH_LBM
        dfixed_equip_mass_dmass_coeff_7 = dfurnishing_wt_dmass_coeff_7 / GRAV_ENGLISH_LBM

        num_flight_attendants = 0.0
        if PAX >= 20.0:
            num_flight_attendants = 1.0
        if PAX >= 51.0:
            num_flight_attendants = 2.0
        if PAX >= 101.0:
            num_flight_attendants = 3.0
        if PAX >= 151.0:
            num_flight_attendants = 4.0
        if PAX >= 201.0:
            num_flight_attendants = 5.0
        if PAX >= 251.0:
            num_flight_attendants = 6.0

        if engine_type is GASPEngineType.TURBOJET:
            doil_per_eng_wt_dFn_SLS = 0.0054
        # elif engine_type is GASPEngineType.TURBOSHAFT or engine_type is GASPEngineType.TURBOPROP:
        #     doil_per_eng_wt_dFn_SLS = 0.0124
        # else:
        #     doil_per_eng_wt_dFn_SLS = 0.062
        else:
            # Other engine types are currently not supported in Aviary
            doil_per_eng_wt_dFn_SLS = 0.0

        dservice_wt_dmass_coeff_8 = 0.0
        if PAX > 9.0:
            dservice_wt_dmass_coeff_8 = PAX * GRAV_ENGLISH_LBM

        dwater_wt_dmass_coeff_9 = 0.0
        if PAX > 19.0:
            dwater_wt_dmass_coeff_9 = (
                PAX + num_pilots + num_flight_attendants) * GRAV_ENGLISH_LBM

        demergency_wt_dmass_coeff_10 = 0.0
        if ~(-1e-5 < mass_coeff[10] < 1e-5):
            demergency_wt_dmass_coeff_10 = GRAV_ENGLISH_LBM

        dcatering_wt_dmass_coeff_11 = 0.0
        if PAX > 19.0:
            dcatering_wt_dmass_coeff_11 = PAX * GRAV_ENGLISH_LBM

        dtrapped_fuel_wt_dmass_coeff_12 = (wing_area**0.5) * fuel_vol_frac / 0.430
        dtrapped_fuel_wt_dwing_area = (
            0.5 * mass_coeff[12] * (wing_area**-0.5) * fuel_vol_frac / 0.430
        )
        dtrapped_fuel_wt_dfuel_vol_frac = mass_coeff[12] * (wing_area**0.5) / 0.430

        if (
            fuel_vol_frac <= 0.075
        ):  # note: this technically creates a discontinuity # won't change
            dtrapped_fuel_wt_dmass_coeff_12 = 0.18 * (wing_area**0.5)
            dtrapped_fuel_wt_dwing_area = (
                0.5 * mass_coeff[12] * 0.18 * (wing_area**-0.5)
            )
            dtrapped_fuel_wt_dfuel_vol_frac = 0.0

        doil_wt_dFnSLS = num_engines * doil_per_eng_wt_dFn_SLS

        duseful_mass_dFn_SLS = doil_wt_dFnSLS / GRAV_ENGLISH_LBM
        duseful_mass_dmass_coeff_8 = dservice_wt_dmass_coeff_8 / GRAV_ENGLISH_LBM
        duseful_mass_dmass_coeff_9 = dwater_wt_dmass_coeff_9 / GRAV_ENGLISH_LBM
        duseful_mass_dmass_coeff_10 = demergency_wt_dmass_coeff_10 / GRAV_ENGLISH_LBM
        duseful_mass_dmass_coeff_11 = dcatering_wt_dmass_coeff_11 / GRAV_ENGLISH_LBM
        duseful_mass_dmass_coeff_12 = dtrapped_fuel_wt_dmass_coeff_12 / GRAV_ENGLISH_LBM
        duseful_mass_dwing_area = dtrapped_fuel_wt_dwing_area / GRAV_ENGLISH_LBM
        duseful_mass_dfuel_vol_frac = dtrapped_fuel_wt_dfuel_vol_frac / GRAV_ENGLISH_LBM

        partials[Aircraft.Design.FIXED_USEFUL_LOAD,
                 Aircraft.Engine.SCALED_SLS_THRUST] = duseful_mass_dFn_SLS
        partials[Aircraft.Design.FIXED_USEFUL_LOAD, Aircraft.Design.EQUIPMENT_MASS_COEFFICIENTS] = np.array(
            [
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                duseful_mass_dmass_coeff_8,
                duseful_mass_dmass_coeff_9,
                duseful_mass_dmass_coeff_10,
                duseful_mass_dmass_coeff_11,
                duseful_mass_dmass_coeff_12,
            ],
            dtype=object,
        )
        partials[Aircraft.Design.FIXED_USEFUL_LOAD,
                 Aircraft.Wing.AREA] = duseful_mass_dwing_area
        partials[
            Aircraft.Design.FIXED_USEFUL_LOAD, Aircraft.Fuel.WING_FUEL_FRACTION
        ] = duseful_mass_dfuel_vol_frac

        partials[Aircraft.Design.FIXED_EQUIPMENT_MASS, Aircraft.Design.EQUIPMENT_MASS_COEFFICIENTS] = np.array(
            [
                dfixed_equip_mass_dmass_coeff_0,
                dfixed_equip_mass_dmass_coeff_1,
                dfixed_equip_mass_dmass_coeff_2,
                dfixed_equip_mass_dmass_coeff_3,
                dfixed_equip_mass_dmass_coeff_4,
                dfixed_equip_mass_dmass_coeff_5,
                dfixed_equip_mass_dmass_coeff_6,
                dfixed_equip_mass_dmass_coeff_7,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
            ],
            dtype=object,
        )
        partials[
            Aircraft.Design.FIXED_EQUIPMENT_MASS, Mission.Design.GROSS_MASS
        ] = dfixed_equip_wt_dgross_wt_initial
        partials[Aircraft.Design.FIXED_EQUIPMENT_MASS,
                 Aircraft.Fuselage.LENGTH] = dfixed_equip_mass_dfus_len
        partials[Aircraft.Design.FIXED_EQUIPMENT_MASS,
                 Aircraft.Wing.SPAN] = dfixed_equip_mass_dwingspan
        partials[Aircraft.Design.FIXED_EQUIPMENT_MASS,
                 Aircraft.Controls.TOTAL_MASS] = dfixed_equip_wt_dcontrol_wt
        partials[
            Aircraft.Design.FIXED_EQUIPMENT_MASS, Aircraft.LandingGear.TOTAL_MASS
        ] = dfixed_equip_wt_dlanding_gear_weight

        partials[Aircraft.Design.FIXED_EQUIPMENT_MASS,
                 Aircraft.Fuselage.PRESSURE_DIFFERENTIAL] = dfixed_equip_mass_dp_diff_fus
        partials[Aircraft.Design.FIXED_EQUIPMENT_MASS,
                 Aircraft.Fuselage.AVG_DIAMETER] = dfixed_equip_mass_dcabin_width

        partials[Aircraft.Design.FIXED_EQUIPMENT_MASS,
                 Aircraft.Wing.AREA] = dfixed_equip_mass_dwing_area
        partials[
            Aircraft.Design.FIXED_EQUIPMENT_MASS, Aircraft.HorizontalTail.AREA
        ] = dfixed_equip_mass_dhtail_area
        partials[
            Aircraft.Design.FIXED_EQUIPMENT_MASS, Aircraft.VerticalTail.AREA
        ] = dfixed_equip_mass_dvtail_area



"""
OpenMDAO component to compute the design mach number and design coefficient of lift,
based on the calculations used in FLOPS and LEAPS 1.0.
"""
import numpy as np
import openmdao.api as om
from openmdao.components.interp_util.interp import InterpND

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


class Design(om.ExplicitComponent):
    """
    Calculates the design mach number and coefficient of lift.

    Based on subroutines MDESN and CLDESN in FLOPS.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.sub_sonic_coeff = [0.029, 0.1843, 57.2958, 1.0 / 10.0]
        self.super_sonic_coeff = [-0.06416, 0.530389, -0.214493, 0.0376684, 1.0 / 3.0]
        self.des_mach_coeff = [0.32, 57.2958, 0.144]

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        # Aircraft design inputs
        add_aviary_input(self, Aircraft.Wing.ASPECT_RATIO, 0.0)
        add_aviary_input(self, Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN, 0.0)
        add_aviary_input(self, Aircraft.Wing.SWEEP, 0.0)
        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD, 0.0)

        # Declare outputs
        add_aviary_output(self, Mission.Design.MACH, 0.0)
        add_aviary_output(self, Mission.Design.LIFT_COEFFICIENT, 0.0)

    def setup_partials(self):
        self.declare_partials(of='*', wrt='*')

    def compute(self, inputs, outputs):
        aviary_options: AviaryValues = self.options['aviary_options']
        AITEK = aviary_options.get_val(Aircraft.Wing.AIRFOIL_TECHNOLOGY)
        VMAX = aviary_options.get_val(Mission.Constraints.MAX_MACH)

        AR, CAM, SW25, TC = inputs.values()

        if TC.real > 0.065:  # subsonic
            a, b, c, d = self.sub_sonic_coeff
            CLDES = (a + b*AR) * np.cos(SW25/c) * (1.0 + d*CAM) / np.sqrt(AR)

        else:  # supersonic
            a, b, c, d, e = self.super_sonic_coeff
            FAR = AR * TC**e
            CLDES = a + b*FAR + c*FAR**2.0 + d*FAR**3.0

        outputs[Mission.Design.LIFT_COEFFICIENT] = CLDES

        if TC.real > 0.065 or VMAX < 1.0:  # subsonic
            TC23 = TC ** (2.0 / 3.0)

            x = np.array([CLDES[0], TC23[0]], dtype=CLDES.dtype)
            ANS1 = CMDEStable.interpolate(x)
            ANS2 = AMDEStable.interpolate(x)

            ANS = ANS1 * (2.0 - AITEK) + ANS2 * (AITEK - 1.0)
            DESM2D = np.sqrt(ANS + 1.0)

        else:  # supersonic
            x = np.array([CLDES[0], TC[0]])
            DESM2D, self.dDESM2D = HSMDEStable.interpolate(x, compute_derivative=True)

        # Calculate value of design Mach number
        a, b, c = self.des_mach_coeff

        DMDSWP = a * (1.0 - np.cos(SW25 / b))
        DMDAR = c / AR

        # Design Mach number
        outputs[Mission.Design.MACH] = DESM2D + DMDSWP + DMDAR

    def compute_partials(self, inputs, partials):
        aviary_options: AviaryValues = self.options['aviary_options']
        AITEK = aviary_options.get_val(Aircraft.Wing.AIRFOIL_TECHNOLOGY)
        VMAX = aviary_options.get_val(Mission.Constraints.MAX_MACH)

        AR, CAM, SW25, TC = inputs.values()

        if TC.real > 0.065:  # subsonic
            a, b, c, d = self.sub_sonic_coeff
            fact1 = a + b * AR
            fact2 = np.cos(SW25/c)
            fact3 = (1.0 + d*CAM)
            fact4 = 1.0 / np.sqrt(AR)
            CLDES = fact1 * fact2 * fact3 * fact4

            # Calculate derivative of design CL wrt Mach, CL, AR, TC, CAM, and SW25
            dCLDES_dAR = fact3 * fact2 * (b * fact4 - fact1 / (2*AR**1.5))
            dCLDES_dTC = 0.0
            dCLDES_dCAM = d * fact1 * fact2 * fact4
            dCLDES_dSW25 = -fact1 * fact3 * np.sin(SW25/c) * fact4/c

        else:  # supersonic
            a, b, c, d, e = self.super_sonic_coeff
            FAR = AR * TC**e
            CLDES = a + b*FAR + c*FAR**2.0 + d*FAR**3.0

            # Calculate derivative of design CL wrt AR, TC, SW25, and CAM
            dFAR_dAR = TC**e
            dFAR_dTC = e * AR * TC**(e - 1.0)
            dCLDES_dFAR = b + 2.0*c*FAR + 3.0*d*FAR**2.0
            dCLDES_dAR = dCLDES_dFAR * dFAR_dAR
            dCLDES_dTC = dCLDES_dFAR * dFAR_dTC
            dCLDES_dSW25 = 0.0
            dCLDES_dCAM = 0.0

        partials[Mission.Design.LIFT_COEFFICIENT, Aircraft.Wing.ASPECT_RATIO] = \
            dCLDES_dAR
        partials[Mission.Design.LIFT_COEFFICIENT, Aircraft.Wing.THICKNESS_TO_CHORD] = \
            dCLDES_dTC
        partials[
            Mission.Design.LIFT_COEFFICIENT, Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN
        ] = dCLDES_dCAM
        partials[Mission.Design.LIFT_COEFFICIENT, Aircraft.Wing.SWEEP] = dCLDES_dSW25

        if TC.real > 0.065 or VMAX < 1.0:  # subsonic
            TC23 = TC ** (2.0 / 3.0)

            x = np.array([CLDES[0], TC23[0]])
            ANS1, dANS1 = CMDEStable.interpolate(x, compute_derivative=True)
            ANS2, dANS2 = AMDEStable.interpolate(x, compute_derivative=True)

            ANS = ANS1 * (2.0 - AITEK) + ANS2 * (AITEK - 1.0)

            dANS = dANS1 * (2.0 - AITEK) + dANS2 * (AITEK - 1.0)
            dANS_dCLDES = dANS[0, 0]
            dANS_dTC23 = dANS[0, 1]

            dDESM2D_dANS = 0.5 / np.sqrt(ANS + 1)

            # wrt CLDES
            dDESM2D_dCLDES = dDESM2D_dANS * dANS_dCLDES

            # wrt AR
            dTC23_dAR = 0.0
            dANS_dAR = dANS_dCLDES * dCLDES_dAR + dANS_dTC23 * dTC23_dAR
            dDESM2D_dAR = dDESM2D_dANS * dANS_dAR

            # wrt TC
            dTC23_dTC = (2.0/3.0) * TC**(-1.0/3.0)
            dANS_dTC = dANS_dCLDES * dCLDES_dTC + dANS_dTC23 * dTC23_dTC
            dDESM2D_dTC = dDESM2D_dANS * dANS_dTC

            # wrt CAM
            dTC23_dCAM = 0.0
            dANS_dCAM = dANS_dCLDES * dCLDES_dCAM + dANS_dTC23 * dTC23_dCAM
            dDESM2D_dCAM = dDESM2D_dANS * dANS_dCAM

            # wrt SW25
            dTC23_dSW25 = 0.0
            dANS_dSW25 = dANS_dCLDES * dCLDES_dSW25 + dANS_dTC23 * dTC23_dSW25
            dDESM2D_dSW25 = dDESM2D_dANS * dANS_dSW25

        else:  # supersonic
            dDESM2D = self.dDESM2D

            # wrt CLDES
            dDESM2D_dCLDES = dDESM2D[0, 0]

            # wrt AR
            dDESM2D_dAR = dDESM2D_dCLDES * dCLDES_dAR

            # wrt TC
            dDESM2D_dTC = dDESM2D[0, 1] + dDESM2D_dCLDES * dCLDES_dTC

            # wrt CAM
            dTC_dCAM = 0.0
            dDESM2D_dCAM = dDESM2D_dCLDES * dCLDES_dCAM + dDESM2D_dTC * dTC_dCAM

            # wrt SW25
            dTC_dSW25 = 0.0
            dDESM2D_dSW25 = dDESM2D_dCLDES * dCLDES_dSW25 + dDESM2D_dTC * dTC_dSW25

        # wrt AR
        a, b, c = self.des_mach_coeff
        dDMDAR_dAR = -c / AR**2
        dDESM_dAR = dDESM2D_dAR + dDMDAR_dAR

        # wrt TC
        dDESM_dTC = dDESM2D_dTC

        # wrt CAM
        dDESM_dCAM = dDESM2D_dCAM

        # wrt SW25
        dDMDSWP_dSW25 = (a / b) * np.sin(SW25 / b)
        dDESM_dSW25 = dDESM2D_dSW25 + dDMDSWP_dSW25

        partials[Mission.Design.MACH, Aircraft.Wing.ASPECT_RATIO] = dDESM_dAR
        partials[Mission.Design.MACH, Aircraft.Wing.THICKNESS_TO_CHORD] = dDESM_dTC
        partials[Mission.Design.MACH, Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN] = \
            dDESM_dCAM
        partials[Mission.Design.MACH, Aircraft.Wing.SWEEP] = dDESM_dSW25


AMDES = np.array(
    [[8003,     .180,   .240,   .300],
     [.10,   -.2080, -.3330, -.4590],
     [.20,   -.2180, -.3430, -.4680],
     [.30,   -.2290, -.3530, -.4780],
     [.40,   -.2420, -.3650, -.4880],
     [.50,   -.2580, -.3770, -.4950],
     [.60,   -.2710, -.3880, -.5070],
     [.70,   -.2940, -.4130, -.5300],
     [.80,   -.3170, -.4310, -.5460]])

CMDES = np.array(
    [[7003,   0.180000,  0.240000,  0.300000],
     [0.100000,  -0.374000, -0.445000, -0.513000],
     [0.200000,  -0.385000, -0.461000, -0.537000],
     [0.300000,  -0.401000, -0.478000, -0.556000],
     [0.400000,  -0.416000, -0.490000, -0.564000],
     [0.500000,  -0.441000, -0.509000, -0.578000],
     [0.600000,  -0.474000, -0.532000, -0.591000],
     [0.700000,  -0.518000, -0.571000, -0.621000]])

HSMDES = np.array(
    [[6003,   0.020000,  0.040000,  0.060000],
     [0.000000,   0.844000,  0.822000,  0.801000],
     [0.100000,   0.836000,  0.815000,  0.793000],
     [0.200000,   0.829000,  0.807000,  0.786000],
     [0.300000,   0.820000,  0.799000,  0.778000],
     [0.400000,   0.811000,  0.791000,  0.770000],
     [0.500000,   0.802000,  0.781000,  0.759000]])

AMDEStable = InterpND(
    method='slinear', points=(AMDES[1:, 0], AMDES[0, 1:]), values=AMDES[1:, 1:],
    extrapolate=True)
CMDEStable = InterpND(
    method='slinear', points=(CMDES[1:, 0], CMDES[0, 1:]), values=CMDES[1:, 1:],
    extrapolate=True)
HSMDEStable = InterpND(
    method='slinear', points=(HSMDES[1:, 0], HSMDES[0, 1:]), values=HSMDES[1:, 1:],
    extrapolate=True)



import unittest

import openmdao.api as om

from aviary.constants import RHO_SEA_LEVEL_ENGLISH
from aviary.mission.gasp_based.phases.v_rotate_comp import VRotateComp
from aviary.variable_info.variables import Aircraft


class TestVRotateComp(unittest.TestCase):

    def test_partials(self):
        prob = om.Problem()

        prob.model.add_subsystem("vrot_comp", VRotateComp(), promotes_inputs=[
                                 "*"], promotes_outputs=["*"])

        prob.setup(force_alloc_complex=True)

        prob.set_val("dV1", val=10, units="kn")
        prob.set_val("dVR", val=5, units="kn")
        prob.set_val(Aircraft.Wing.AREA, val=1370, units="ft**2")
        prob.set_val("rho", val=RHO_SEA_LEVEL_ENGLISH, units="slug/ft**3")
        prob.set_val("CL_max", val=2.1886, units="unitless")
        prob.set_val("mass", val=175_000, units="lbm")

        prob.run_model()

        # print(prob.get_val("Vrot", units="kn"))

        prob.check_partials(method='cs')



from copy import deepcopy
import pkg_resources
import unittest

import openmdao.api as om

from aviary.interface.methods_for_level2 import AviaryProblem
from aviary.subsystems.subsystem_builder_base import SubsystemBuilderBase
from aviary.interface.default_phase_info.height_energy import phase_info as ph_in
from aviary.variable_info.variables import Aircraft
from openmdao.utils.testing_utils import use_tempdirs


class WingWeightSubsys(om.ExplicitComponent):

    def setup(self):

        self.add_input(Aircraft.Engine.MASS, 1.0, units='lbm')
        self.add_output(Aircraft.Canard.ASPECT_RATIO, 1.0, units='unitless')
        self.add_output('Tail', 1.0, units='unitless')

        self.declare_partials(Aircraft.Canard.ASPECT_RATIO,
                              Aircraft.Engine.MASS, val=2.0)
        self.declare_partials('Tail', Aircraft.Engine.MASS, val=0.7)

    def compute(self, inputs, outputs):
        outputs[Aircraft.Canard.ASPECT_RATIO] = 2.0 * inputs[Aircraft.Engine.MASS]
        outputs['Tail'] = 0.7 * inputs[Aircraft.Engine.MASS]


class WingWeightBuilder(SubsystemBuilderBase):
    """
    Prototype of a subsystem that overrides an aviary internally computed var.
    """

    def __init__(self, name='wing_weight'):
        super().__init__(name)

    def build_post_mission(self, aviary_inputs):
        '''
        Build an OpenMDAO system for the pre-mission computations of the subsystem.

        Returns
        -------
        pre_mission_sys : openmdao.core.System
            An OpenMDAO system containing all computations that need to happen in
            the pre-mission part of the Aviary problem. This
            includes sizing, design, and other non-mission parameters.
        '''
        wing_group = om.Group()
        wing_group.add_subsystem("aerostructures", WingWeightSubsys(),
                                 promotes_inputs=["aircraft:*"],
                                 promotes_outputs=[Aircraft.Canard.ASPECT_RATIO,
                                                   ('Tail',
                                                    Aircraft.Canard.WETTED_AREA_SCALER)])
        return wing_group


@use_tempdirs
class PreMissionGroupTest(unittest.TestCase):

    def test_post_mission_promotion(self):
        phase_info = deepcopy(ph_in)
        phase_info['post_mission'] = {}
        phase_info['post_mission']['include_landing'] = False
        phase_info['post_mission']['external_subsystems'] = [
            WingWeightBuilder(name="wing_external")
        ]

        prob = AviaryProblem()

        csv_path = pkg_resources.resource_filename(
            "aviary", "models/test_aircraft/aircraft_for_bench_GwFm.csv")

        prob.load_inputs(csv_path, phase_info)


# Preprocess inputs
        # prob.check_and_preprocess_inputs()

        prob.add_pre_mission_systems()

        prob.add_phases()

        prob.add_post_mission_systems()

        # Link phases and variables
        prob.link_phases()

        prob.add_driver("SLSQP")

        prob.add_design_variables()

        prob.add_objective(objective_type="mass", ref=-1e5)

        prob.setup()


if __name__ == '__main__':
    unittest.main()



"""
This script is designed to test run scripts located in a specified directory.
It uses Python's unittest framework to dynamically generate test cases for each
script that begins with 'run_' and ends with '.py'. The script ensures that
each run script executes without errors within a 10-second timeout. If a script
runs longer than 10 seconds, it's still considered a pass unless other errors occur.
"""

import os
import subprocess
import unittest
from pathlib import Path


class RunScriptTest(unittest.TestCase):
    """
    A test case class that uses unittest to run and test scripts with a timeout.

    ...

    Attributes
    ----------
    base_directory : str
        The base directory where the run scripts are located.
    run_files : list
        A list of paths to run scripts found in the base directory.

    Methods
    -------
    setUpClass()
        Class method to find all run scripts before tests are run.
    find_run_files(base_dir)
        Finds and returns all run scripts in the specified directory.
    run_script(script_path)
        Attempts to run a script with a timeout and handles errors.
    test_run_scripts()
        Generates a test for each run script with a timeout.
    """

    @classmethod
    def setUpClass(cls):
        """
        Class method to set up the test case class by finding all run scripts.

        This method is called once before starting the tests and is used to
        populate the 'run_files' attribute with a list of run scripts.
        """
        base_directory = os.path.join(os.path.dirname(
            os.path.dirname(os.path.abspath(__file__))), ".")
        cls.run_files = cls.find_run_files(base_directory)

    @staticmethod
    def find_run_files(base_dir):
        """
        Find and return a list of run scripts in the specified directory.

        Parameters
        ----------
        base_dir : str
            The directory to search for run scripts.

        Returns
        -------
        list
            A list of pathlib.Path objects pointing to the run scripts.
        """
        run_files = []
        for root, _, files in os.walk(base_dir):
            for file in files:
                if file.startswith('run_') and file.endswith('.py'):
                    run_files.append(Path(root) / file)
        return run_files

    def run_script(self, script_path):
        """
        Attempt to run a script with a 10-second timeout and handle errors.

        Parameters
        ----------
        script_path : pathlib.Path
            The path to the script to be run.

        Raises
        ------
        Exception
            Any exception other than ImportError or TimeoutExpired that occurs while running the script.
        """
        proc = subprocess.Popen(['python', script_path],
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        proc.wait()
        (stdout, stderr) = proc.communicate()

        if proc.returncode != 0:
            if 'ImportError' in str(stderr):
                self.skipTest(f"Skipped {script_path.name} due to ImportError")
            else:
                raise Exception(
                    f"Error running {script_path.name}:\n{stderr.decode('utf-8')}")

    def test_run_scripts(self):
        """
        Test each run script to ensure it executes without error.

        This method generates a subtest for each script in 'run_files'.
        Each script is tested to ensure it runs without errors.
        """
        for script_path in self.run_files:
            with self.subTest(script=script_path.name):
                self.run_script(script_path)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python

import argparse
import re

import numpy as np

from enum import Enum
from pathlib import Path

from aviary.api import NamedValues
from aviary.utils.csv_data_file import write_data_file
from aviary.utils.functions import get_path


class CodeOrigin(Enum):
    FLOPS = 'FLOPS'
    GASP = 'GASP'


allowed_headers = {
    'altitude': 'Altitude',
    'alt': 'Altitude',
    'alpha': 'Angle of Attack',
    'mach': 'Mach',
    'delflp': 'Flap Deflection',
    'cltot': 'CL',
    'cl': 'CL',
    'cd': 'CD',
    'hob': 'Hob',
    'del_cl': 'Delta CL',
    'del_cd': 'Delta CD'
}


def AeroDataConverter(input_file=None, output_file=None, data_format=None):
    data_format = CodeOrigin(data_format)
    data_file = get_path(input_file)
    if not output_file:
        if data_format is CodeOrigin.GASP:
            # Default output file name is same location and name as input file, with
            # '_aviary' appended to filename
            path = input_file.parents[0]
            name = input_file.name
            suffix = input_file.suffix
            output_file = path / (name + '_aviary' + suffix)
        elif data_format is CodeOrigin.FLOPS:
            # Default output file name is same location and name as input file, with
            # '_aviary' appended to filename
            path = input_file.parents[0]
            name = input_file.stem
            suffix = input_file.suffix
            file1 = path / name + '_aviary_CDI' + suffix
            file2 = path / name + '_aviary_CD0' + suffix
            output_file = [file1, file2]

    stamp = f'# {data_format.value}-derived aerodynamics data converted from {data_file.name}'

    if data_format is CodeOrigin.GASP:
        data, comments = _load_gasp_aero_table(data_file)
        comments = [stamp] + comments

        write_data_file(output_file, data, comments, include_timestamp=True)
    elif data_format is CodeOrigin.FLOPS:
        if type(output_file) is not list:
            # if only one filename is given, split into two
            path = output_file.parents[0]
            name = output_file.stem
            suffix = output_file.suffix
            file1 = path / (name + '_CDi' + suffix)
            file2 = path / (name + '_CD0' + suffix)
            output_file = [file1, file2]

        lift_drag_data, lift_drag_comments, \
            zero_lift_drag_data, zero_lift_drag_comments = _load_flops_aero_table(
                data_file)

        # write lift-dependent drag file
        lift_drag_comments = [stamp] + lift_drag_comments
        write_data_file(output_file[0], lift_drag_data,
                        lift_drag_comments, include_timestamp=True)

        # write zero-lift drag file
        zero_lift_drag_comments = [stamp] + zero_lift_drag_comments
        write_data_file(output_file[1], zero_lift_drag_data,
                        zero_lift_drag_comments, include_timestamp=True)


def _load_flops_aero_table(filepath: Path):
    """Load an aero table in FLOPS format"""

    def _read_line(line_count, comments):
        line = file_contents[line_count].strip()
        items = re.split(r'[\s]*\s', line)
        if items[0] == '#':
            comments.append(line)
            nonlocal offset
            offset += 1
            try:
                items = _read_line(line_count+1, comments)
            except IndexError:
                return
        else:
            # try to convert line to float
            try:
                items = [float(var) for var in items]
            # data contains things other than floats
            except (ValueError):
                raise ValueError(
                    f'Non-numerical value found in data file <{filepath.name}> on '
                    f'line {str(line_count)}')

        return items

    lift_drag = []
    lift_drag_data = NamedValues()
    lift_drag_comments = []
    zero_lift_drag = []
    zero_lift_drag_data = NamedValues()
    zero_lift_drag_comments = []

    file_contents = []
    with open(filepath, 'r') as reader:
        for line in reader:
            file_contents.append(line)

    offset = 0
    # these are not needed, we can determine the length of data vectors directly
    lift_drag_mach_count, cl_count = _read_line(0 + offset, lift_drag_comments)
    lift_drag_machs = _read_line(1 + offset, lift_drag_comments)
    cls = _read_line(2 + offset, lift_drag_comments)
    lift_drag = []
    for i in range(len(lift_drag_machs)):
        drag = _read_line(3 + i + offset, lift_drag_comments)
        if len(drag) == len(cls):
            lift_drag.append(drag)
        else:
            raise ValueError('Number of data points provided for '
                             f'lift-dependent drag at Mach {lift_drag_machs[i]} '
                             'does not match number of CLs provided '
                             f'(FLOPS aero data file {filepath.name})')
    if len(lift_drag) != len(lift_drag_machs):
        raise ValueError('Number of data rows provided for lift-dependent drag does '
                         'not match number of Mach numbers provided (FLOPS aero data '
                         f'file {filepath.name})')
    offset = offset + i
    # these are not needed, we can determine the length of data vectors directly
    altitude_count, zero_lift_mach_count = _read_line(
        4 + offset, zero_lift_drag_comments)
    altitudes = _read_line(5 + offset, zero_lift_drag_comments)
    zero_lift_machs = _read_line(6 + offset, zero_lift_drag_comments)
    for i in range(len(altitudes)):
        drag = _read_line(7 + i + offset, zero_lift_drag_comments)
        if len(drag) == len(zero_lift_machs):
            zero_lift_drag.append(drag)
        else:
            raise ValueError('Number of data points provided for '
                             f'zero-lift drag at altitude {altitudes[i]} '
                             'does not match number of Machs provided '
                             f'(FLOPS aero data file {filepath.name})')
    if len(zero_lift_drag) != len(altitudes):
        raise ValueError('Number of data rows provided for zero-lift drag does '
                         'not match number of altitudes provided (FLOPS aero data '
                         f'file {filepath.name})')

    cl, mach = np.meshgrid(cls, lift_drag_machs)
    lift_drag_data.set_val('Mach', mach.flatten(), 'unitless')
    lift_drag_data.set_val('Lift Coefficient', cl.flatten(), 'unitless')
    lift_drag_data.set_val('Lift-Dependent Drag Coefficient',
                           np.array(lift_drag).flatten(), 'unitless')

    mach, altitude = np.meshgrid(zero_lift_machs, altitudes)
    zero_lift_drag_data.set_val('Altitude', altitude.flatten(), 'ft')
    zero_lift_drag_data.set_val('Mach', mach.flatten(), 'unitless')
    zero_lift_drag_data.set_val('Zero-Lift Drag Coefficient',
                                np.array(zero_lift_drag).flatten(), 'unitless')

    return lift_drag_data, lift_drag_comments, zero_lift_drag_data, zero_lift_drag_comments


def _load_gasp_aero_table(filepath: Path):
    """Load an aero table in GASP format"""
    data = NamedValues()
    raw_data = []
    comments = []
    variables = []
    units = []
    read_header = True
    read_units = False
    with open(filepath, 'r') as reader:
        for line_count, line in enumerate(reader):
            # ignore empty lines
            if not line or line.strip() == ['']:
                continue
            if line[0] == '#':
                line = line.strip('# ').strip()
                if read_header:
                    items = re.split(r'[\s]*\s', line)
                    if all(name.lower() in allowed_headers for name in items):
                        variables = [name for name in items]
                        read_header = False
                        read_units = True
                    else:
                        if line:
                            comments.append(line)
                else:
                    if read_units:
                        items = re.split(r'[\s]*\s', line)
                        for item in items:
                            item = item.strip('()')
                            if item == '-':
                                item = 'unitless'
                            units.append(item)
                continue
            else:
                # this is data
                items = re.split(r'[\s]*\s', line.strip())
                # try to convert line to float
                try:
                    line_data = [float(var) for var in items]
                # data contains things other than floats
                except (ValueError):
                    raise ValueError(
                        f'Non-numerical value found in data file <{filepath.name}> on '
                        f'line {str(line_count)}')
                else:
                    raw_data.append(line_data)

    raw_data = np.array(raw_data)

    # translate raw data into NamedValues object
    for idx, var in enumerate(variables):
        if var.lower() in allowed_headers:
            data.set_val(allowed_headers[var.lower()], raw_data[:, idx], units[idx])

    return data, comments


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Converts FLOPS- or GASP-formatted '
                                     'aero data files into Aviary csv format.\n')
    parser.add_argument('input_file', type=str,
                        help='path to engine deck file to be converted')
    parser.add_argument('output_file', type=str,
                        help='path to file where new converted data will be written')
    parser.add_argument('data_format', type=str, choices=[origin.value for origin in CodeOrigin],
                        help='data format used by input_file')

    args = parser.parse_args()

    AeroDataConverter(**vars(args))



"""
This file contains functions needed to run Aviary using the Level 1 interface.
"""
import os

import openmdao.api as om
from aviary.variable_info.enums import AnalysisScheme
from aviary.interface.methods_for_level2 import AviaryProblem


def run_aviary(aircraft_filename, phase_info, optimizer=None,
               analysis_scheme=AnalysisScheme.COLLOCATION, objective_type=None,
               record_filename='dymos_solution.db', restart_filename=None, max_iter=50,
               run_driver=True, make_plots=True, phase_info_parameterization=None,
               optimization_history_filename=None):
    """
    Run the Aviary optimization problem for a specified aircraft configuration and mission.

    This function creates an instance of the AviaryProblem class using provided phase information,
    mission, and mass methods. It processes aircraft and options data from the given aircraft filename,
    checks for clashing inputs, and sets up pre- and post-mission systems. The optimization problem is formulated,
    initialized with guesses, and run to find a solution.

    Parameters
    ----------
    aircraft_filename : str
        Filename from which to load the aircraft and options data.
    phase_info : dict
        Information about the phases of the mission.
    optimizer : str
        The optimizer to use.
    analysis_scheme : AnalysisScheme, optional
        The analysis scheme to use, defaults to AnalysisScheme.COLLOCATION.
    objective_type : str, optional
        Type of the optimization objective.
    record_filename : str, optional
        Filename for recording the solution, defaults to 'dymos_solution.db'.
    restart_filename : str, optional
        Filename to use for restarting the optimization, if applicable.
    max_iter : int, optional
        Maximum number of iterations for the optimizer, defaults to 50.
    run_driver : bool, optional
        If True, the driver will be run, defaults to True.
    make_plots : bool, optional
        If True, generate plots during the optimization, defaults to True.
    phase_info_parameterization : function, optional
        Additional information to parameterize the phase_info object based on
        desired cruise altitude and Mach.

    Returns
    -------
    AviaryProblem
        The AviaryProblem instance after running the optimization problem.

    Notes
    -----
    The function allows for user overrides on aircraft and options data.
    It raises warnings or errors if there are clashing user inputs.
    Users can modify or add methods to alter the Aviary problem's behavior.
    """

    # Build problem
    prob = AviaryProblem(analysis_scheme)

    # Load aircraft and options data from user
    # Allow for user overrides here
    prob.load_inputs(aircraft_filename, phase_info)

    # Preprocess inputs
    prob.check_and_preprocess_inputs()

    prob.add_pre_mission_systems()

    prob.add_phases(phase_info_parameterization=phase_info_parameterization)

    prob.add_post_mission_systems()

    # Link phases and variables
    prob.link_phases()

    prob.add_driver(optimizer, max_iter=max_iter)

    prob.add_design_variables()

    # Load optimization problem formulation
    # Detail which variables the optimizer can control
    prob.add_objective(objective_type=objective_type)

    prob.setup()

    prob.set_initial_guesses()

    prob.failed = prob.run_aviary_problem(
        record_filename, restart_filename=restart_filename, run_driver=run_driver, make_plots=make_plots, optimization_history_filename=optimization_history_filename)

    return prob


def run_level_1(
    input_deck,
    outdir='output',
    optimizer='SNOPT',
    phase_info=None,
    n2=False,
    max_iter=50,
    analysis_scheme=AnalysisScheme.COLLOCATION,
):
    '''
    This file enables running aviary from the command line with a user specified input deck.
    usage: aviary run_mission [input_deck] [opt_args]
    '''

    kwargs = {
        'max_iter': max_iter,
    }

    if analysis_scheme is AnalysisScheme.SHOOTING:
        kwargs['analysis_scheme'] = AnalysisScheme.SHOOTING
        kwargs['run_driver'] = False
    #     kwargs['optimizer'] = 'IPOPT'
    # else:
    kwargs['optimizer'] = optimizer

    prob = run_aviary(input_deck, phase_info, **kwargs)

    if n2:
        outfile = os.path.join(outdir, "n2.html")
        if outdir != '':
            os.makedirs(outdir, exist_ok=True)
        om.n2(
            prob,
            outfile=outfile,
            show_browser=False,
        )

    return prob


def _setup_level1_parser(parser):
    def_outdir = os.path.join(os.getcwd(), "output")
    parser.add_argument(
        'input_deck', metavar='indeck', type=str, nargs=1, help='Name of vehicle input deck file'
    )
    parser.add_argument(
        "-o", "--outdir", default=def_outdir, help="Directory to write outputs"
    )
    parser.add_argument(
        "--optimizer",
        type=str,
        default='SNOPT',
        help="Name of optimizer",
        choices=("SNOPT", "IPOPT", "SLSQP", "None")
    )
    parser.add_argument(
        "--phase_info",
        type=str,
        default=None,
        help="Path to phase info file"
    )
    parser.add_argument("--n2", action="store_true",
                        help="Generate an n2 diagram after the analysis")
    parser.add_argument(
        "--max_iter",
        type=int,
        default=50,
        help="maximum number of iterations")
    parser.add_argument(
        "--shooting",
        action="store_true",
        help="Use shooting instead of collocation",
    )


def _exec_level1(args, user_args):
    if args.shooting:  # For future use
        analysis_scheme = AnalysisScheme.SHOOTING
    else:
        analysis_scheme = AnalysisScheme.COLLOCATION

    if args.optimizer == 'None':
        args.optimizer = None

    # check if args.input_deck is a list, if so, use the first element
    if isinstance(args.input_deck, list):
        args.input_deck = args.input_deck[0]

    prob = run_level_1(
        input_deck=args.input_deck,
        outdir=args.outdir,
        optimizer=args.optimizer,
        phase_info=args.phase_info,
        n2=args.n2,
        max_iter=args.max_iter,
        analysis_scheme=analysis_scheme,
    )



from aviary.variable_info.variables import Aircraft as _Aircraft
from aviary.variable_info.variables import Mission as _Mission


def compare_inner_classes(class1, class2, show_all=False):
    """
    Compare two nested class hierarchies and return a set of shared inner-classes.

    Summary:
    This function takes in two classes that both are part of variable hierarchies and may contain
    inner-classes or variables with string-named values. This function compares those two classes 
    and returns a set of inner classes that have the same name in both inputted classes. It will throw 
    an error if the two classes have the same variable with a different string-named value.

    Parameters
    ----------
    class1 : class
        A class that is all or part of a variable hierarchy. This can be the top-level 
        class in the hierarchy, or any inner class nested at any depth within that top-level class.
    class2 : class
        A class that is all or part of a variable hierarchy. This can be the top-level
        class in the hierarchy, or any inner class nested at any depth within that top-level class.
    show_all : bool, optional
        Flag to tell the function to return the sets of variables and inner classes for
        each provided class.

    Returns
    -------
    overlapping_inner_classes : set of strings
        A set of string names of all the inner classes which are common between the two
        input classes.
    class1_vars_set : set of strings, optional
        Set of string names of the variables belonging to class1, optional return based 
        on show_all flag.
    class2_vars_set : set of strings, optional
        Set of string names of the variables belonging to class2, optional return based
        on show_all flag.
    class1_inner_classes_set : set of strings, optional
        Set of the string names of inner classes belonging to class1, optional return based
        on show_all flag.
    class2_inner_classes_set : set of strings, optional
        Set of the string names of inner classes belonging to class2, optional return based
        on show_all flag.

    Raises
    ------
    ValueError
        If the two input classes both have a variable with the same variable name but 
        different string-named value.
    """
    class1_vars_inner_classes = vars(class1)
    class2_vars_inner_classes = vars(class2)

    class1_vars = []
    class1_inner_classes = []
    class2_vars = []
    class2_inner_classes = []

    # separate out a list of string names of the variables belonging to class1, and the inner classes belonging to class1
    for key in class1_vars_inner_classes.keys():
        # just checks if it is a class
        if type(class1_vars_inner_classes[key]) == type(class1):
            class1_inner_classes.append(key)
        elif (type(class1_vars_inner_classes[key]) == str) and not (key == '__module__') and not (key == '__doc__'):
            class1_vars.append(key)

    # separate out a list of string names of the variables belonging to class2, and the inner classes belonging to class2
    for key in class2_vars_inner_classes.keys():
        if type(class2_vars_inner_classes[key]) == type(class2):
            class2_inner_classes.append(key)
        elif (type(class2_vars_inner_classes[key]) == str) and not (key == '__module__') and not (key == '__doc__'):
            class2_vars.append(key)

    class1_vars_set = set(class1_vars)
    class2_vars_set = set(class2_vars)
    # get set of the variables in the provided classes that have the same name
    overlapping_vars = class1_vars_set & class2_vars_set
    for var in overlapping_vars:  # go through overlapping variables and check that they have the same value associated with them
        value1 = getattr(class1, var)  # value of the variable in the first class
        value2 = getattr(class2, var)  # value of the variable in the second class

        if value1 != value2:
            raise ValueError(
                f"You have attempted to merge two variable hierarchies together that have the same variable with a different string name associated to it. The offending variable is '{var}'. In '{class1.__qualname__}' it has a value of '{value1}' and in '{class2.__qualname__}' it has a value of '{value2}'.")

    class1_inner_classes_set = set(class1_inner_classes)
    class2_inner_classes_set = set(class2_inner_classes)
    overlapping_inner_classes = class1_inner_classes_set & class2_inner_classes_set

    if show_all:
        return (overlapping_inner_classes, class1_vars_set, class2_vars_set, class1_inner_classes_set, class2_inner_classes_set)

    else:
        return (overlapping_inner_classes)


def recursive_comparison(overlapping_inner_classes, outer_class_a, outer_class_b):
    """
    Recursively compares all inner classes to an infinite depth and identifies mismatched string-named values.

    For all of the inner class names provided in overlapping_inner_classes this function calls compare_inner_classes
    and compares those inner classes recursively until it reaches the full depth to which outer_class_a and
    outer_class_b have any inner classes in common.

    Parameters
    ----------
    overlapping_inner_classes : set of strings
        This is a set of strings where each string is the name of an inner class that outer_class_a
        and outer_class_b have in common.
    outer_class_a : class
        A class that is all or part of a variable hierarchy. This can be the top-level class in the 
        hierarchy, or any inner class nested at any depth within that top-level class.
    outer_class_b : class
        A class that is all or part of a variable hierarchy. This can be the top-level class in the 
        hierarchy, or any inner class nested at any depth within that top-level class.

    Returns
    -------
    None

    Exceptions
    ----------
    No exceptions explicitly raised by this function, although called functions may raise exceptions.
    """
    for overlapping_class_name in overlapping_inner_classes:
        overlapping_inner_class_a = getattr(outer_class_a, overlapping_class_name)
        overlapping_inner_class_b = getattr(outer_class_b, overlapping_class_name)
        overlapping_second_inner_classes = compare_inner_classes(
            overlapping_inner_class_a, overlapping_inner_class_b)
        recursive_comparison(overlapping_second_inner_classes,
                             overlapping_inner_class_a, overlapping_inner_class_b)


def compare_hierarchies_to_merge(hierarchies_to_merge):
    """
    Compares variable hierarchies to ensure there are no string-valued variable conflicts.

    For all the variable hierarchies provided in hierarchies_to_merge this function compares each
    hierarchy with every other hierarchy as well as the Aviary core aircraft and mission hierarchies
    to ensure there are no string-valued variable conflicts within the same class or inner class to
    and infinite depth.

    Parameters
    ----------
    hierarchies_to_merge : list of classes
        This is a list of variable hierarchy classes which will be compared for merge compatibility
        with one another.

    Returns
    -------
    None

    Raises
    ------
    No explicit exceptions are raised by this function, although called functions may raise exceptions.
    """

    for hierarchy in hierarchies_to_merge:
        # check if hierarchy has developed conflicts with original hierarchy
        if issubclass(hierarchy, _Aircraft):
            overlap = compare_inner_classes(hierarchy, _Aircraft)
            recursive_comparison(overlap, hierarchy, _Aircraft)

        # check if hierarchy has developed conflicts with original hierarchy
        if issubclass(hierarchy, _Mission):
            overlap = compare_inner_classes(hierarchy, _Mission)
            recursive_comparison(overlap, hierarchy, _Mission)

        for hierarchy2 in hierarchies_to_merge:
            if hierarchy != hierarchy2:
                overlap = compare_inner_classes(hierarchy, hierarchy2)
                recursive_comparison(overlap, hierarchy, hierarchy2)



import subprocess
import unittest
from pathlib import Path
import shutil

import pkg_resources
from openmdao.utils.testing_utils import use_tempdirs
from aviary.interface.download_models import get_model, save_file


@use_tempdirs
class CommandEntryPointsTestCases(unittest.TestCase):

    def run_and_test_hangar(self, filenames, out_dir=''):
        # tests that the commands return an exit code of 0 and that the files are generated
        if isinstance(filenames, str):
            filenames = [filenames]
        cmd = ['aviary', 'hangar'] + filenames

        if out_dir:
            cmd += ['-o', out_dir]
            out_dir = Path(out_dir)
        else:
            out_dir = Path.cwd() / 'aviary_models'

        try:
            output = subprocess.check_output(cmd)
            for filename in filenames:
                path = out_dir / filename.split('/')[-1]
                self.assertTrue(path.exists())
        except subprocess.CalledProcessError as err:
            self.fail(f"Command '{cmd}' failed.  Return code: {err.returncode}")

    def test_single_file_without_path(self):
        filename = 'turbofan_22k.deck'
        self.run_and_test_hangar(filename)

    def test_single_file_with_path(self):
        filename = 'engines/turbofan_22k.deck'
        self.run_and_test_hangar(filename)

    def test_multiple_files(self):
        filenames = ['small_single_aisle_GwGm.dat', 'small_single_aisle_GwGm.csv']
        self.run_and_test_hangar(filenames)

    def test_folder(self):
        filename = 'engines'
        self.run_and_test_hangar(filename)

    def test_single_file_custom_outdir(self):
        filename = 'small_single_aisle_GwGm.csv'
        out_dir = '~/test_hangar'
        self.run_and_test_hangar(filename, out_dir)
        shutil.rmtree(out_dir)

    def test_expected_path(self):
        filename = f'test_aircraft/converter_configuration_test_data_GwGm.dat'
        aviary_path = get_model('converter_configuration_test_data_GwGm.dat')

        expected_path = pkg_resources.resource_filename('aviary',
                                                        'models/test_aircraft/converter_configuration_test_data_GwGm.dat')
        self.assertTrue(str(aviary_path) == expected_path)


if __name__ == "__main__":
    unittest.main()



import dymos as dm
import openmdao.api as om
from dymos.utils.misc import _unspecified
from openmdao.core.component import Component

from aviary.utils.aviary_values import AviaryValues

import dymos as dm

from dymos.utils.misc import _unspecified

import openmdao.api as om

from aviary.variable_info.core_promotes import core_mission_inputs
from aviary.variable_info.variable_meta_data import _MetaData

# ---------------------------
# Helper functions for setting up inputs/outputs in components
# ---------------------------


def add_aviary_input(comp, varname, val=None, units=None, desc=None, shape_by_conn=False, meta_data=_MetaData):
    '''
    This function provides a clean way to add variables from the
    variable hierarchy into components as Aviary inputs. It takes
    the standard OpenMDAO inputs of the variable's name, initial
    value, units, and description, as well as the component which
    the variable is being added to.
    '''
    meta = meta_data[varname]
    if units:
        input_units = units
    else:
        # units of None are overwritten with defaults. Overwriting units with None is
        # unecessary as it will cause errors down the line if the default is not already
        # None
        input_units = meta['units']
    if desc:
        input_desc = desc
    else:
        input_desc = meta['desc']
    if val is None:
        val = meta['default_value']
    comp.add_input(varname, val=val, units=input_units,
                   desc=input_desc, shape_by_conn=shape_by_conn)


def add_aviary_output(comp, varname, val, units=None, desc=None, shape_by_conn=False, meta_data=_MetaData):
    '''
    This function provides a clean way to add variables from the
    variable hierarchy into components as Aviary outputs. It takes
    the standard OpenMDAO inputs of the variable's name, initial
    value, units, and description, as well as the component which
    the variable is being added to.
    '''
    meta = meta_data[varname]
    if units:
        output_units = units
    else:
        # units of None are overwritten with defaults. Overwriting units with None is
        # unecessary as it will cause errors down the line if the default is not already
        # None
        output_units = meta['units']
    if desc:
        output_desc = desc
    else:
        output_desc = meta['desc']
    comp.add_output(varname, val=val, units=output_units,
                    desc=output_desc, shape_by_conn=shape_by_conn)


def override_aviary_vars(group, aviary_inputs: AviaryValues,
                         manual_overrides=None, external_overrides=None):
    '''
    This function provides the capability to override output variables
    with variables from the aviary_inputs input. The user can also
    optionally provide the names of variables that they would like to
    override manually. (Manual overriding is simply suppressing the
    promotion of the variable to make way for another output variable
    of the same name, or to create an unconnected input elsewhere.)
    '''
    def name_filter(name):
        return "aircraft:" in name or "mission:" in name

    if not manual_overrides:
        manual_overrides = []

    if not external_overrides:
        external_overrides = []

    # first need to make a list of all the inputs that anyone needs
    # so that we can keep track of any unclaimed inputs
    all_inputs = set()  # use a set to avoid duplicates
    for system in group.system_iter():
        meta = system.get_io_metadata(iotypes=("input",))
        in_var_names = meta.keys()
        for name in in_var_names:
            all_inputs.add(name)

    overridden_outputs = []
    external_overridden_outputs = []
    for comp in group.system_iter(typ=Component):
        # get a list of the variables to use
        out_var_names = list(filter(name_filter, comp.get_io_metadata(
            iotypes=("output",), return_rel_names=False)))
        # get a list of the metadata associated with each variable
        out_var_metadata = comp.get_io_metadata(
            iotypes=("output",), return_rel_names=False)
        in_var_names = filter(name_filter, comp.get_io_metadata(iotypes=("input", )))

        comp_promoted_outputs = []

        for abs_name in out_var_names:
            name = out_var_metadata[abs_name]['prom_name']

            if abs_name in manual_overrides:
                # These are handled outside of this function.
                continue

            elif name in external_overrides:

                # Overridden variables are given a new name
                comp_promoted_outputs.append((name, f"EXTERNAL_OVERRIDE:{name}"))
                external_overridden_outputs.append(name)

                continue  # don't promote it

            elif name in aviary_inputs:
                if name in all_inputs:
                    val, units = aviary_inputs.get_item(name)
                    group.set_input_defaults(name, val=val, units=units)

                    # Overridden variables are given a new name
                    comp_promoted_outputs.append((name, f"AUTO_OVERRIDE:{name}"))
                    overridden_outputs.append(name)

                continue  # don't promote it

            # This variable is not overriden, so the output is promoted.
            comp_promoted_outputs.append(name)

        # note: Always promoting all inputs into the "global" namespace
        # so its VERY important that we enforce all inputs names exist in the master
        # variable list
        rel_path = comp.pathname[len(group.pathname):].lstrip(".")
        if "." in rel_path:
            # comp is in a subgroup. We must find it.
            sub_path = ".".join(rel_path.split(".")[:-1])
            sub = group._get_subsystem(sub_path)
            sub.promotes(comp.name, inputs=in_var_names, outputs=comp_promoted_outputs)
        else:
            group.promotes(comp.name, inputs=in_var_names, outputs=comp_promoted_outputs)

    if overridden_outputs:
        print("\nThe following variables have been overridden:")
        for prom_name in sorted(overridden_outputs):
            print(f"  '{prom_name}")

    if external_overridden_outputs:
        print("\nThe following variables have been overridden by an external subsystem:")
        for prom_name in sorted(external_overridden_outputs):
            print(f"  '{prom_name}")

    return overridden_outputs


def setup_trajectory_params(
    model: om.Group, traj: dm.Trajectory, aviary_variables: AviaryValues, phases=['climb', 'cruise', 'descent'],
    variables_to_add=None, meta_data=_MetaData, external_parameters={},
):
    '''
    This function smoothly sorts through the aviary variables which
    are being used in the trajectory, and for the variables which are
    not options it adds them as a parameter of the trajectory.
    '''
    # TODO: make this automated as part of the trajectory subclass
    if variables_to_add is None:
        variables_to_add = sorted(core_mission_inputs)

    for key in sorted(variables_to_add):
        meta = meta_data[key]

        if not meta['option']:
            val = meta['default_value']
            if val is None:
                val = _unspecified
            units = meta['units']

            if key in aviary_variables:
                try:
                    val = aviary_variables.get_val(key, units)
                except TypeError:
                    val = aviary_variables.get_val(key)

                    # TODO temp line to ignore dynamic mission variables, will not work
                    #      if names change to 'dynamic:mission:*'
            if ':' not in key:
                continue

            # TODO: make a trajectory subclass that does this
            traj.add_parameter(
                key,
                opt=False,
                units=units,
                val=val,
                static_target=True,
                targets={phase_name: [key] for phase_name in phases})

            model.promotes('traj', inputs=[(f'parameters:{key}', key)])

    # Step 1: Initialize a dictionary to hold parameters and their associated phases
    parameters_with_phases = {}

    # Step 2: Loop through external_parameters to populate the dictionary
    for phase_name, parameter_dict in external_parameters.items():
        for key in parameter_dict.keys():
            if key not in parameters_with_phases:
                parameters_with_phases[key] = []
            parameters_with_phases[key].append(phase_name)

    # Step 3: Loop through the collected parameters and call traj.add_parameter
    for key, phases in parameters_with_phases.items():
        # Assuming the kwargs are the same for shared parameters
        kwargs = external_parameters[phases[0]][key]
        targets = {phase: [key] for phase in phases}
        traj.add_parameter(
            key,
            **kwargs,
            targets=targets
        )
        model.promotes('traj', inputs=[(f'parameters:{key}', key)])

    return traj


def get_units(key, meta_data=None) -> str:
    """
    Returns the units for the specified variable as defined in the MetaData.

    Parameters
    ----------
    key: str
        Name of the variable
    meta_data : dict
        Dictionary containing metadata for the variable. If None, Aviary's built-in
        metadata will be used.
    """
    if meta_data is None:
        meta_data = _MetaData

    return meta_data[key]['units']



import unittest

import openmdao.api as om
from parameterized import parameterized

from aviary.subsystems.mass.flops_based.wing_detailed import \
    DetailedWingBendingFact
from aviary.utils.test_utils.variable_test import assert_match_varnames
from aviary.validation_cases.validation_tests import (flops_validation_test,
                                                      get_flops_case_names,
                                                      get_flops_inputs,
                                                      print_case)
from aviary.variable_info.variables import Aircraft, Mission


class DetailedWingBendingFactTest(unittest.TestCase):

    def setUp(self):
        self.prob = om.Problem()


class DetailedWingBendingTest(unittest.TestCase):

    def setUp(self):
        self.prob = om.Problem()

    # Skip model that doesn't use detailed wing.
    @parameterized.expand(get_flops_case_names(omit=['LargeSingleAisle2FLOPS', 'LargeSingleAisle2FLOPSalt']),
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        self.prob.model.add_subsystem(
            "wing",
            DetailedWingBendingFact(aviary_options=get_flops_inputs(case_name)),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )
        self.prob.setup(check=False, force_alloc_complex=True)

        flops_validation_test(
            prob,
            case_name,
            input_keys=[Aircraft.Wing.LOAD_PATH_SWEEP_DIST,
                        Aircraft.Wing.THICKNESS_TO_CHORD_DIST,
                        Aircraft.Wing.CHORD_PER_SEMISPAN_DIST,
                        Mission.Design.GROSS_MASS,
                        Aircraft.Engine.POD_MASS,
                        Aircraft.Wing.ASPECT_RATIO,
                        Aircraft.Wing.ASPECT_RATIO_REF,
                        Aircraft.Wing.STRUT_BRACING_FACTOR,
                        Aircraft.Wing.AEROELASTIC_TAILORING_FACTOR,
                        Aircraft.Engine.WING_LOCATIONS,
                        Aircraft.Wing.THICKNESS_TO_CHORD,
                        Aircraft.Wing.THICKNESS_TO_CHORD_REF],
            output_keys=[Aircraft.Wing.BENDING_FACTOR,
                         Aircraft.Wing.ENG_POD_INERTIA_FACTOR],
            method='fd',
            atol=1e-3,
            rtol=1e-5)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


if __name__ == "__main__":
    unittest.main()



import unittest
import os

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import (assert_check_partials,
                                         assert_near_equal)

from aviary.mission.gasp_based.ode.constraints.speed_constraints import \
    SpeedConstraints
from aviary.variable_info.variables import Dynamic


class SpeedConstraintTestCase1(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group",
            SpeedConstraints(num_nodes=3, EAS_target=229, mach_cruise=0.8),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults("EAS", np.array([229, 229, 229]), units="kn")
        self.prob.model.set_input_defaults(
            Dynamic.Mission.MACH, np.array([0.6, 0.6, 0.6]), units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        tol = 1e-6
        self.prob.run_model()

        assert_near_equal(
            self.prob["speed_constraint"],
            np.array([[0, -45.8], [0, -45.8], [0, -45.8]]),
            tol,
        )

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


class SpeedConstraintTestCase2(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group",
            SpeedConstraints(num_nodes=3, EAS_target=229, mach_cruise=0.8),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults("EAS", np.array([229, 229, 229]), units="kn")
        self.prob.model.set_input_defaults(
            Dynamic.Mission.MACH, np.array([0.9, 0.9, 0.9]), units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        tol = 1e-6
        self.prob.run_model()

        assert_near_equal(
            self.prob["speed_constraint"],
            np.array([[0, 22.9], [0, 22.9], [0, 22.9]]),
            tol,
        )

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


if __name__ == "__main__":
    unittest.main()



import unittest

import openmdao.api as om
from parameterized import parameterized

from aviary.subsystems.mass.flops_based.unusable_fuel import (
    AltUnusableFuelMass, TransportUnusableFuelMass)
from aviary.utils.test_utils.variable_test import assert_match_varnames
from aviary.validation_cases.validation_tests import (Version,
                                                      flops_validation_test,
                                                      get_flops_case_names,
                                                      get_flops_inputs,
                                                      print_case)
from aviary.variable_info.variables import Aircraft

# TODO: update non-transport tests to remove these variables
_wing_ref_area = 'aircraft:wing:dimensions:area'
_engine_count_factor = 'aircraft:propulsion:control:engine_count_factor'
_max_scaled_thrust = 'aircraft:propulsion:dimensions:max_scaled_thrust'
_total_fuel_capacity = 'aircraft:fuel:mass:total_capacity'
_density_ratio = 'aircraft:fuel:mass:density_ratio'
_tank_count = 'aircraft:fuel:control:tank_count'
_total_fuel_vol = 'TBD:total_fuel_vol:'
_unusable_fuel_mass = 'TBD:unusable_fuel'


class TransportUnusableFuelMassTest(unittest.TestCase):
    '''
    Tests transport/GA unusable fuel mass calculation.
    '''

    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(get_flops_case_names(),
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob
        flops_inputs = get_flops_inputs(case_name)

        prob.model.add_subsystem(
            'unusable_fuel',
            TransportUnusableFuelMass(aviary_options=flops_inputs),
            promotes_outputs=['*'],
            promotes_inputs=['*']
        )

        prob.setup(check=False, force_alloc_complex=True)

        flops_validation_test(
            prob,
            case_name,
            input_keys=[Aircraft.Fuel.UNUSABLE_FUEL_MASS_SCALER,
                        Aircraft.Fuel.DENSITY_RATIO,
                        Aircraft.Fuel.TOTAL_CAPACITY,
                        Aircraft.Propulsion.TOTAL_SCALED_SLS_THRUST,
                        Aircraft.Wing.AREA],
            output_keys=[  # Aircraft.Fuel.TOTAL_VOLUME,
                Aircraft.Fuel.UNUSABLE_FUEL_MASS],
            version=Version.TRANSPORT,
            tol=5e-4,
            excludes=['size_prop.*'])

    def test_IO(self):
        assert_match_varnames(self.prob.model)


class AltUnusableFuelMassTest(unittest.TestCase):
    '''
    Tests alternate unusable fuel mass calculation.
    '''

    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(get_flops_case_names(),
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        prob.model.add_subsystem(
            'unusable_fuel',
            AltUnusableFuelMass(aviary_options=get_flops_inputs(case_name)),
            promotes_outputs=['*'],
            promotes_inputs=['*']
        )

        prob.setup(check=False, force_alloc_complex=True)

        flops_validation_test(
            prob,
            case_name,
            input_keys=[Aircraft.Fuel.UNUSABLE_FUEL_MASS_SCALER,
                        Aircraft.Fuel.TOTAL_CAPACITY],
            output_keys=[  # Aircraft.Fuel.TOTAL_VOLUME,
                Aircraft.Fuel.UNUSABLE_FUEL_MASS],
            version=Version.ALTERNATE)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


if __name__ == '__main__':
    unittest.main()



"""
Test file to test the outputs, derivatives, and IO of each sample component/group.
The name of this file needs to start with 'test' so that the testflo command will
find and run the file.
"""

import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import (assert_check_partials,
                                         assert_near_equal)

from aviary import constants
from aviary.mission.flops_based.phases.simplified_takeoff import (
    FinalTakeoffConditions, StallSpeed, TakeoffGroup)
from aviary.variable_info.variables import Aircraft, Dynamic, Mission


class StallSpeedTest(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "comp",
            StallSpeed(),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults("mass", val=181200.0, units="lbm")  # check
        self.prob.model.set_input_defaults(
            "rho", val=constants.RHO_SEA_LEVEL_METRIC, units="kg/m**3"
        )  # check
        self.prob.model.set_input_defaults(
            "planform_area", val=1370.0, units="ft**2"
        )  # check (this is the reference wing area)
        self.prob.model.set_input_defaults(
            'Cl_max', val=2.0000, units='unitless')  # check

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        self.prob.run_model()

        tol = 1e-5

        assert_near_equal(self.prob["v_stall"], 71.90002053, tol)  # not actual value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(
            partial_data, atol=1e-12, rtol=1e-12
        )  # check the partial derivatives


class FinalConditionsTest(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "comp", FinalTakeoffConditions(num_engines=2), promotes=["*"]
        )

        self.prob.model.set_input_defaults(
            "v_stall", val=100, units="m/s"
        )  # not actual value
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=181200.0, units="lbm"
        )  # check
        self.prob.model.set_input_defaults(
            Mission.Takeoff.FUEL_SIMPLE, val=577, units="lbm"
        )  # check
        self.prob.model.set_input_defaults(
            "rho", val=constants.RHO_SEA_LEVEL_ENGLISH, units="slug/ft**3"
        )  # check
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1370.0, units="ft**2"
        )  # check
        self.prob.model.set_input_defaults(
            Mission.Takeoff.LIFT_COEFFICIENT_MAX, val=2.0000, units='unitless')  # check
        self.prob.model.set_input_defaults(
            Mission.Design.THRUST_TAKEOFF_PER_ENG, val=28928.0, units="lbf")  # check
        self.prob.model.set_input_defaults(
            Mission.Takeoff.LIFT_OVER_DRAG, val=17.354, units='unitless')  # check

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        self.prob.run_model()

        tol = 1e-5

        assert_near_equal(
            self.prob[Mission.Takeoff.GROUND_DISTANCE], 6637.65417226, tol
        )  # ft (not actual value)
        # m/s (not actual value)
        assert_near_equal(self.prob[Mission.Takeoff.FINAL_VELOCITY], 123.09, tol)
        assert_near_equal(
            self.prob[Mission.Takeoff.FINAL_MASS], 180623.0, tol
        )  # lbm (not actual value)
        assert_near_equal(
            self.prob[Mission.Takeoff.FINAL_ALTITUDE], 35, tol)  # ft

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


class TakeoffGroupTest(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group_example", TakeoffGroup(num_engines=2), promotes=["*"]
        )

        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=181200.0, units="lbm"
        )  # check
        self.prob.model.set_input_defaults(
            Mission.Takeoff.FUEL_SIMPLE, val=577, units="lbm"
        )  # check
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1370.0, units="ft**2"
        )  # check
        self.prob.model.set_input_defaults(
            Mission.Takeoff.LIFT_COEFFICIENT_MAX, val=2.0000, units='unitless')  # check
        self.prob.model.set_input_defaults(
            Mission.Design.THRUST_TAKEOFF_PER_ENG, val=28928.0, units="lbf")  # check
        self.prob.model.set_input_defaults(
            Mission.Takeoff.LIFT_OVER_DRAG, val=17.354, units='unitless')  # check
        self.prob.model.set_input_defaults(
            Dynamic.Mission.ALTITUDE, val=0, units="ft")  # check

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        self.prob.run_model()

        tol = 1e-5

        assert_near_equal(
            self.prob[Mission.Takeoff.GROUND_DISTANCE], 6637.58951504, tol
        )  # ft (not actual value)
        assert_near_equal(
            self.prob[Mission.Takeoff.FINAL_VELOCITY], 88.49655173, tol
        )  # m/s (not actual value)
        assert_near_equal(
            self.prob[Mission.Takeoff.FINAL_MASS], 180623.0, tol
        )  # lbm (not actual value)
        assert_near_equal(
            self.prob[Mission.Takeoff.FINAL_ALTITUDE], 35, tol)  # ft

        partial_data = self.prob.check_partials(
            out_stream=None, excludes=['*.USatm'], method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


if __name__ == "__main__":
    unittest.main()



import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class CableSize(om.ExplicitComponent):

    def initialize(self):

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        add_aviary_input(self, Aircraft.Engine.WING_LOCATIONS, val=0.35)

        add_aviary_input(self, Aircraft.Wing.SPAN, val=128)

        add_aviary_input(self, Aircraft.Fuselage.AVG_DIAMETER, val=10)

        add_aviary_output(self, Aircraft.Electrical.HYBRID_CABLE_LENGTH, val=0)

        self.declare_partials(
            Aircraft.Electrical.HYBRID_CABLE_LENGTH,
            [
                Aircraft.Engine.WING_LOCATIONS,
                Aircraft.Wing.SPAN,
            ],
        )

        self.declare_partials(
            Aircraft.Electrical.HYBRID_CABLE_LENGTH,
            Aircraft.Fuselage.AVG_DIAMETER,
            val=2.)

    def compute(self, inputs, outputs):

        eng_span_frac = inputs[Aircraft.Engine.WING_LOCATIONS]
        wingspan = inputs[Aircraft.Wing.SPAN]
        cabin_width = inputs[Aircraft.Fuselage.AVG_DIAMETER]

        outputs[Aircraft.Electrical.HYBRID_CABLE_LENGTH] = (
            eng_span_frac * wingspan + 2.0 * cabin_width
        )

    def compute_partials(self, inputs, J):

        eng_span_frac = inputs[Aircraft.Engine.WING_LOCATIONS]
        wingspan = inputs[Aircraft.Wing.SPAN]

        J[
            Aircraft.Electrical.HYBRID_CABLE_LENGTH,
            Aircraft.Engine.WING_LOCATIONS,
        ] = wingspan
        J[Aircraft.Electrical.HYBRID_CABLE_LENGTH, Aircraft.Wing.SPAN] = eng_span_frac



import numpy as np
import openmdao.api as om
from dymos.models.atmosphere.atmos_1976 import USatm1976Comp

from aviary.mission.gasp_based.ode.base_ode import BaseODE
from aviary.mission.gasp_based.ode.params import ParamPort
from aviary.mission.gasp_based.ode.rotation_eom import RotationEOM
from aviary.variable_info.enums import AnalysisScheme
from aviary.variable_info.variables import Aircraft, Dynamic, Mission


class RotationODE(BaseODE):
    """ODE for takeoff rotation.

    This phase spans the time from when the aircraft is touching the runway but has
    begun to rotate to liftoff.
    """

    def setup(self):
        nn = self.options["num_nodes"]
        analysis_scheme = self.options["analysis_scheme"]
        aviary_options = self.options['aviary_options']
        core_subsystems = self.options['core_subsystems']

        # TODO: paramport
        self.add_subsystem("params", ParamPort(), promotes=["*"])

        self.add_subsystem(
            "USatm", USatm1976Comp(
                num_nodes=nn), promotes_inputs=[
                ("h", Dynamic.Mission.ALTITUDE)], promotes_outputs=[
                "rho", ("sos", Dynamic.Mission.SPEED_OF_SOUND), ("temp", Dynamic.Mission.TEMPERATURE), ("pres", Dynamic.Mission.STATIC_PRESSURE), "viscosity"], )

        self.add_flight_conditions(nn)

        kwargs = {'num_nodes': nn, 'aviary_inputs': aviary_options,
                  'method': 'low_speed'}
        for subsystem in core_subsystems:
            system = subsystem.build_mission(**kwargs)
            if system is not None:
                self.add_subsystem(subsystem.name,
                                   system,
                                   promotes_inputs=subsystem.mission_inputs(**kwargs),
                                   promotes_outputs=subsystem.mission_outputs(**kwargs))

        if analysis_scheme is AnalysisScheme.SHOOTING:
            alpha_comp = om.ExecComp(
                'alpha=rotation_rate*(t_curr-start_rotation)+alpha_init',
                alpha=dict(val=0., units='deg'),
                rotation_rate=dict(val=10.0/3.0, units='deg/s'),
                t_curr=dict(val=0., units='s'),
                start_rotation=dict(val=0., units='s'),
                alpha_init=dict(val=0., units='deg'),
            )
            alpha_comp_inputs = ["rotation_rate", "t_curr", "start_rotation",
                                 ("alpha_init", Aircraft.Wing.INCIDENCE)]
            self.add_subsystem("alpha_comp",
                               alpha_comp,
                               promotes_inputs=alpha_comp_inputs,
                               promotes_outputs=["alpha"],
                               )

        self.add_subsystem("eoms", RotationEOM(
            num_nodes=nn, analysis_scheme=analysis_scheme), promotes=["*"])

        ParamPort.set_default_vals(self)
        self.set_input_defaults("t_init_flaps", val=47.5)
        self.set_input_defaults("t_init_gear", val=37.3)
        self.set_input_defaults("alpha", val=np.ones(nn), units="deg")
        self.set_input_defaults(Dynamic.Mission.FLIGHT_PATH_ANGLE,
                                val=np.zeros(nn), units="deg")
        self.set_input_defaults(Dynamic.Mission.ALTITUDE, val=np.zeros(nn), units="ft")
        self.set_input_defaults(Dynamic.Mission.VELOCITY, val=np.zeros(nn), units="kn")
        self.set_input_defaults("t_curr", val=np.zeros(nn), units="s")
        self.set_input_defaults('aero_ramps.flap_factor:final_val', val=1.)
        self.set_input_defaults('aero_ramps.gear_factor:final_val', val=1.)
        self.set_input_defaults('aero_ramps.flap_factor:initial_val', val=1.)
        self.set_input_defaults('aero_ramps.gear_factor:initial_val', val=1.)



import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary.subsystems.geometry.gasp_based.strut import StrutGeom
from aviary.variable_info.options import get_option_defaults
from aviary.variable_info.variables import Aircraft


class SizeGroupTestCase1(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem("strut", StrutGeom(
            aviary_options=get_option_defaults()), promotes=["*"])

        self.prob.model.set_input_defaults(
            Aircraft.Strut.AREA_RATIO, val=.2, units=None
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=150, units="ft**2"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Strut.ATTACHMENT_LOCATION, val=1.0
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, val=10.0, units="ft"
        )  # not actual GASP value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4
        assert_near_equal(self.prob["strut_y"], 0.5, tol)  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Strut.LENGTH], 10.9658561, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Strut.CHORD], 1.36788226, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


if __name__ == "__main__":
    unittest.main()
